"""Top level CoCiP classes and methods."""

from __future__ import annotations

import logging
import warnings
from typing import Any, NoReturn, Sequence, overload

import numpy as np
import pandas as pd
import xarray as xr
from overrides import overrides

from pycontrails.core import met_var
from pycontrails.core.fleet import Fleet
from pycontrails.core.flight import Flight
from pycontrails.core.met import MetDataset
from pycontrails.core.models import Model, interpolate_met
from pycontrails.core.vector import GeoVectorDataset, VectorDataDict
from pycontrails.datalib import ecmwf, gfs
from pycontrails.models import sac, tau_cirrus
from pycontrails.models.aircraft_performance import AircraftPerformance
from pycontrails.models.cocip import (
    contrail_properties,
    radiative_forcing,
    radiative_heating,
    wake_vortex,
    wind_shear,
)
from pycontrails.models.cocip.cocip_params import CocipFlightParams
from pycontrails.models.cocip.output import flight_summary
from pycontrails.models.emissions.emissions import Emissions
from pycontrails.physics import geo, thermo, units

logger = logging.getLogger(__name__)


class Cocip(Model):
    r"""Contrail Cirrus Prediction Model (CoCiP).

    Published by Ulrich Schumann *et. al.*
    (`DLR Institute of Atmospheric Physics <https://www.dlr.de/pa/en/>`_)
    in :cite:`schumannContrailCirrusPrediction2012`, :cite:`schumannParametricRadiativeForcing2012`.

    Parameters
    ----------
    met : MetDataset
        Pressure level dataset containing :attr:`met_variables` variables.
        See *Notes* for variable names by data source.
    rad : MetDataset
        Single level dataset containing top of atmosphere radiation fluxes.
        See *Notes* for variable names by data source.
    params : dict[str, Any], optional
        Override Cocip model parameters with dictionary.
        See :class:`CocipFlightParams` for model parameters.
    **params_kwargs
        Override Cocip model parameters with keyword arguments.
        See :class:`CocipFlightParams` for model parameters.

    Notes
    -----
    **Inputs**

    The required meteorology variables depend on the data source (e.g. ECMWF, GFS).

    See :attr:`met_variables` and :attr:`rad_variables` for the list of required variables
    to the ``met`` and ``rad`` inputs, respectively.
    When an item in one of these arrays is a tuple, variable keys depend on data source.

    The current list of required variables (labelled by ``"standard_name"``):

    .. list-table:: Variable keys for pressure level data
        :header-rows: 1

        * - Parameter
          - ECMWF
          - GFS
        * - Air Temperature
          - ``air_temperature``
          - ``air_temperature``
        * - Specific Humidity
          - ``specific_humidity``
          - ``specific_humidity``
        * - Eastward wind
          - ``eastward_wind``
          - ``eastward_wind``
        * - Northward wind
          - ``northward_wind``
          - ``northward_wind``
        * - Vertical velocity
          - ``lagrangian_tendency_of_air_pressure``
          - ``lagrangian_tendency_of_air_pressure``
        * - Ice water content
          - ``specific_cloud_ice_water_content``
          - ``ice_water_mixing_ratio``
        * - Geopotential
          - ``geopotential``
          - ``geopotential_height``

    .. list-table:: Variable keys for single-level radiation data
        :header-rows: 1

        * - Parameter
          - ECMWF
          - GFS
        * - Top solar radiation
          - ``top_net_solar_radiation``
          - ``toa_upward_shortwave_flux``
        * - Top thermal radiation
          - ``top_net_thermal_radiation``
          - ``toa_upward_longwave_flux``

    **Modifications**

    This implementation differs from original CoCiP (Fortran) implementation in a few places:

    - This model uses aircraft performance and emissions models to calculate nvPM, fuel flow,
      and overall propulsion efficiency, if not already provided.
    - As described in :cite:`teohAviationContrailClimate2022`, this implementation sets
      the initial ice particle activation rate to be a function of
      the difference between the ambient temperature and the critical SAC threshold temperature.
      See :func:`sac.T_critical_sac`.
    - Isobaric heat capacity calculation.
      The original model uses a constant value of 1004 :math:`J \ kg^{-1} \ K^{-1}`,
      whereas this model calculates isobaric heat capacity as a function of specific humidity.
      See :func:`thermo.c_pm`.
    - Solar direct radiation.
      The original algorithm uses ECMWF radiation variable `tisr` (top incident solar radiation)
      as solar direct radiation value.
      This implementation calculates the theoretical solar direct radiation at
      any arbitrary point in the atmosphere.
      See :func:`geo.solar_direct_radiation`.
    - Segment angle.
      The segment angle calculations for flights and contrail segments have been updated
      to use more precise spherical geometry instead of a triangular approximation.
      As the triangle approaches zero, the two calculations agree.
      See :func:`geo.segment_angle`.
    - Integration.
      This implementation consistently uses left-Riemann sums
      in the time integration of contrail segments.
    - Segment length ratio.
      Instead of taking a geometric mean between contrail segments before/after advection,
      a simple ratio is computed.
      See :func:`contrail_properties.segment_length_ratio`.
    - Segment energy flux.
      This implementation does not average spatially contiguous contrail segments when calculating
      the mean energy flux for the segment of interest.
      See :func:`contrail_properties.mean_energy_flux_per_m`.

    This implementation is regression tested against
    results from :cite:`teohAviationContrailClimate2022`.
    See `tests/benchmark/north-atlantic-study/validate.py`.

    **Outputs**

    NaN values may appear in model output. Specifically, `np.nan` values are used to indicate:

    - Flight waypoint or contrail waypoint is not contained with the :attr:`met` domain.
    - The variable was NOT computed during the model evaluation. For example, at flight waypoints
      not producing any persistent contrails, "radiative" variables (`rsr`, `olr`, `rf_sw`,
      `rf_lw`, `rf_net`) are not computed. Consequently, the corresponding values in the output
      of :meth:`eval` are NaN. One exception to this rule is found on `ef` (energy forcing)
      `contrail_age` predictions. For these two "cumulative" variables, waypoints not producing
      any persistent contrails are assigned 0 values.

    References
    ----------
    - :cite:`schumannDeterminationContrailsSatellite1990`
    - :cite:`schumannContrailCirrusPrediction2010`
    - :cite:`voigtInsituObservationsYoung2010`
    - :cite:`schumannPotentialReduceClimate2011a`
    - :cite:`schumannContrailCirrusPrediction2012`
    - :cite:`schumannParametricRadiativeForcing2012`
    - :cite:`schumannContrailsVisibleAviation2012`
    - :cite:`schumannEffectiveRadiusIce2011`
    - :cite:`schumannDehydrationEffectsContrails2015`
    - :cite:`teohMitigatingClimateForcing2020`
    - :cite:`schumannAviationContrailCirrus2021`
    - :cite:`schumannAirTrafficContrail2021`
    - :cite:`teohAviationContrailClimate2022`

    See Also
    --------
    :class:`CocipFlightParams`
    :mod:`wake_vortex`
    :mod:`contrail_properties`
    :mod:`radiative_forcing`
    :mod:`humidity_scaling`
    :class:`Emissions`
    :mod:`sac`
    :mod:`tau_cirrus`
    """

    __slots__ = (
        "rad",
        "contrail",
        "contrail_dataset",
        "contrail_list",
        "timesteps",
        "_sac_flight",
        "_downwash_flight",
        "_downwash_contrail",
    )

    name = "cocip"
    long_name = "Contrail Cirrus Prediction Model"
    default_params = CocipFlightParams
    met_variables = (
        met_var.AirTemperature,
        met_var.SpecificHumidity,
        met_var.EastwardWind,
        met_var.NorthwardWind,
        met_var.VerticalVelocity,
        (ecmwf.SpecificCloudIceWaterContent, gfs.CloudIceWaterMixingRatio),
        (met_var.Geopotential, met_var.GeopotentialHeight),
    )

    #: Required single-level top of atmosphere radiation variables.
    #: Variable keys depend on data source (e.g. ECMWF, GFS).
    rad_variables = (
        (ecmwf.TopNetSolarRadiation, gfs.TOAUpwardShortwaveRadiation),
        (ecmwf.TopNetThermalRadiation, gfs.TOAUpwardLongwaveRadiation),
    )

    #: Minimal set of met variables needed to run the model after pre-processing.
    #: The intention here is that ``geopotential`` and ``ciwc`` are unnecessary after
    #: ``tau_cirrus`` has already been calculated.
    processed_met_variables = (
        met_var.AirTemperature,
        met_var.SpecificHumidity,
        met_var.EastwardWind,
        met_var.NorthwardWind,
        met_var.VerticalVelocity,
        tau_cirrus.TauCirrus,
    )

    #: Additional met variables used to support outputs
    optional_met_variables = ((ecmwf.CloudAreaFractionInLayer, gfs.TotalCloudCoverIsobaric),)

    #: Met data is not optional
    met: MetDataset
    met_required = True

    #: Radiation data formatted as a MetDataset at a single pressure level [-1]
    rad: MetDataset

    #: Last Flight modeled in :meth:`eval`
    source: Flight | Fleet

    #: List of GeoVectorDataset contrail objects - one for each timestep
    contrail_list: list[GeoVectorDataset]

    #: Contrail evolution output from model. Set to None when no contrails are formed.
    contrail: pd.DataFrame | None

    #: xr.Dataset reporesentation of contrail evolution.
    contrail_dataset: xr.Dataset | None

    #: Array of np.datetime64 time steps for contrail evolution
    timesteps: np.ndarray

    #: Parallel copy of flight waypoints after SAC filter applied
    _sac_flight: Flight

    #: Parallel copy of flight waypoints after wake vortex downwash applied
    _downwash_flight: Flight

    #: GeoVectorDataset representation of :attr:`downwash_flight`
    _downwash_contrail: GeoVectorDataset

    def __init__(
        self,
        met: MetDataset,
        rad: MetDataset,
        params: dict[str, Any] = {},
        **params_kwargs: Any,
    ) -> None:
        # call Model init
        super().__init__(met, params=params, **params_kwargs)

        shift_radiation_time = self.params["shift_radiation_time"]
        met, rad = process_met_datasets(met, rad, shift_radiation_time)

        self.met = met
        self.rad = rad

        # initialize outputs to None
        self.contrail = None
        self.contrail_dataset = None

    # ----------
    # Public API
    # ----------

    @overload
    def eval(self, source: Flight, **params: Any) -> Flight:
        ...

    @overload
    def eval(self, source: Sequence[Flight], **params: Any) -> list[Flight]:
        ...

    # This is only included for type consistency with parent. This will raise.
    @overload
    def eval(self, source: None = None, **params: Any) -> NoReturn:
        ...

    def eval(
        self,
        source: Flight | Sequence[Flight] | None = None,
        **params: Any,
    ) -> Flight | list[Flight] | NoReturn:
        """Run CoCiP simulation on flight.

        Simulates the formation and evolution of contrails from a Flight
        using the contrail cirrus prediction model (CoCiP) from Schumann (2012)
        :cite:`schumannContrailCirrusPrediction2012`.

        .. versionchanged:: 0.25.11

            Previously, any waypoint not surviving the wake vortex downwash phase of CoCiP
            was assigned a nan-value in the ``ef`` array within the model
            output. This is no longer the case. Instead, energy forcing is set to 0.0
            for all waypoints which fail to produce persistent contrails. In particular,
            nan values in the ``ef`` array are only used to indicate an
            out-of-met-domain waypoint. The same convention is now used for output variables
            ``contrail_age`` and ``cocip`` as well.

        .. versionchanged::0.26.0

            :attr:`met` and :attr:`rad` down-selection is now handled automatically.

        Parameters
        ----------
        source : Flight | Sequence[Flight]
            Input Flight(s) to model.
        **params : Any
            Overwrite model parameters before eval.

        Returns
        -------
        :class : Flight | list[Flight]
            Flight(s) with updated Contrail data. The model parameter "verbose_outputs"
            determines the variables on the return flight object.

        Raises
        ------
        ValueError
            Raises when ``source`` is None.
            Raises when the met variable "specific_humidity" is improperly scaled.
        KeyError
            Raises when the variable "tau_cirrus" is not in the met dataset.

        References
        ----------
        - :cite:`schumannContrailCirrusPrediction2012`
        """

        self.update_params(params)
        self.set_source(source)
        self.source = self.require_source_type(Flight)
        return_flight_list = isinstance(self.source, Fleet) and isinstance(source, Sequence)

        self._calc_timesteps()

        # Downselect met for CoCiP initialization
        # We only need to buffer in the negative vertical direction,
        # which is the positive direction for level
        logger.debug("Downselect met for Cocip initialization")
        buffer = 0, self.params["met_level_buffer"][1]
        met = self.source.downselect_met(self.met, copy=False, level_buffer=buffer)

        # Prepare flight for model
        self._process_flight(met)

        # Save humidity scaling type to output attrs
        # NOTE: Do this after _process_flight because that method automatically
        # broadcasts all numeric source params.
        if self.params["humidity_scaling"] is not None:
            for k, v in self.params["humidity_scaling"].description.items():
                self.source.attrs[f"humidity_scaling_{k}"] = v

        if isinstance(self.source, Fleet):
            label = f"fleet of {self.source.n_flights} flights"
        else:
            label = f"flight {self.source['flight_id'][0]}"

        logger.debug("Finding initial linear contrails with SAC")
        self._find_initial_contrail_regions()

        if not self._sac_flight:
            logger.debug("No linear contrails formed by %s", label)
            return self._fill_empty_flight_results(return_flight_list)

        self._simulate_wake_vortex_downwash(met)

        self._find_initial_persistent_contrails(met)

        if not self._downwash_flight:
            logger.debug("No persistent contrails formed by flight %s", label)
            return self._fill_empty_flight_results(return_flight_list)

        self.contrail_list = []
        self._simulate_contrail_evolution()

        self._cleanup_indices()

        if not self.contrail_list:
            logger.debug("No contrails formed by %s", label)
            return self._fill_empty_flight_results(return_flight_list)

        logger.debug("Complete contrail simulation for %s", label)

        self._bundle_results()

        if return_flight_list:
            return self.source.to_flight_list()  # type: ignore[attr-defined]

        return self.source

    def output_flight_statistics(self) -> pd.Series:
        """Calculate summary flight statistics for evaluated flight.

        Returns
        -------
        pd.Series
            Summary statistics as a :class:`pandas.Series`

        Raises
        ------
        RuntimeError
            If :meth:`eval` has not been called.
        """
        if "ef" not in self.source:
            raise RuntimeError(
                "Model must be evaluated with `Cocip.eval` before flight statistics are available"
            )

        return flight_summary.flight_statistics(self.source, self.contrail)

    def _calc_timesteps(self) -> None:
        """Calculate :attr:`timesteps`."""
        if isinstance(self.source, Fleet):
            # time not sorted in Fleet instance
            tmin = self.source["time"].min()
            tmax = self.source["time"].max()
        else:
            tmin = self.source["time"][0]
            tmax = self.source["time"][-1]

        tmin = pd.to_datetime(tmin)
        tmax = pd.to_datetime(tmax)
        dt = pd.to_timedelta(self.params["dt_integration"])

        t_start = tmin.ceil(dt)
        t_end = tmax.floor(dt) + self.params["max_age"] + dt
        self.timesteps = np.arange(t_start, t_end, dt)

    # -------------
    # Model Methods
    # -------------

    def _process_flight(self, met: MetDataset) -> None:
        """Prepare :attr:`self.source` for use in model eval.

        Missing flight values should be prefilled before calling this method.
        This method modifies :attr:`self.source`.

        ..versionchanged:: 0.35.2

            No longer broadcast all numeric source params. Instead, numeric
            source params can be accessed with :meth:`Flight.get_data_or_attr`.

        Parameters
        ----------
        met : MetDataset
            Meteorology data

        See Also
        --------
        :method:`Flight.resample_and_fill`
        """
        logger.debug("Pre-processing flight parameters")

        # STEP 1: Check for core columns
        # Attach level and altitude to avoid some redundancy
        self.source.setdefault("altitude", self.source.altitude)
        self.source.setdefault("level", self.source.level)
        self.source.setdefault("air_pressure", self.source.air_pressure)

        core_columns = ["longitude", "latitude", "altitude", "time"]
        for col in core_columns:
            if np.isnan(self.source[col]).any():
                raise ValueError(
                    f"Parameter `flight` must not contain NaN values in {col} field."
                    "Call method `resample_and_fill` to clean up flight trajectory."
                )

        # STEP 2: Check for waypoints
        # Note: Fleet instances always have a waypoint column
        if "waypoint" not in self.source:
            # Give flight a range index
            self.source["waypoint"] = np.arange(self.source.size)
        elif not isinstance(self.source, Fleet):
            # If self.source is a usual Flight (not the subclass Fleet)
            # and has "waypoint" data, we want to give a warning if the waypoint
            # data has an usual index. CoCiP uses the waypoint for its continuity
            # convention, so the waypoint data is critical.
            if not np.all(np.diff(self.source["waypoint"]) == 1):
                raise ValueError(
                    "Found non-sequential waypoints in flight key 'waypoint'. "
                    "The CoCiP algorithm requires flight data key 'waypoint' "
                    "to contain sequential waypoints if defined."
                )

        # STEP 3: Test met domain for some overlap
        # We attach the intersection to the source.
        # This is used in the function _fill_empty_flight_results
        intersection = self.source.coords_intersect_met(met)
        self.source["_met_intersection"] = intersection
        logger.debug(
            "Fraction of flight waypoints intersecting met domain: %s / %s",
            intersection.sum(),
            self.source.size,
        )
        if not intersection.any():
            raise ValueError(
                "No intersection between flight waypoints and met domain. "
                "Rerun Cocip with `met` overlapping flight."
            )

        # STEP 4: Begin met interpolation
        # Unfortunately we use both "u_wind" and "eastward_wind" to refer to the
        # same variable, so the logic gets a bit more complex.
        scale_humidity = (self.params["humidity_scaling"] is not None) and (
            "specific_humidity" not in self.source
        )
        variables = {
            "air_temperature": "air_temperature",
            "specific_humidity": "specific_humidity",
            "eastward_wind": "u_wind",
            "northward_wind": "v_wind",
        }
        for met_variable, fl_variable in variables.items():
            interpolate_met(met, self.source, met_variable, fl_variable, **self.interp_kwargs)

        if scale_humidity:
            self.params["humidity_scaling"].eval(self.source, copy_source=False)

        # if humidity_scaling isn't defined, add rhi to source for verbose_outputs
        elif self.params["verbose_outputs"]:
            self.source["rhi"] = thermo.rhi(
                self.source["specific_humidity"],
                self.source["air_temperature"],
                self.source.air_pressure,
            )

        # Cache extra met properties for post-analysis
        if self.params["verbose_outputs"]:
            interpolate_met(met, self.source, "tau_cirrus", **self.interp_kwargs)

            # handle ECMWF/GFS ciwc variables
            if (key := "specific_cloud_ice_water_content") in met:
                interpolate_met(met, self.source, key, **self.interp_kwargs)
            elif (key := "ice_water_mixing_ratio") in met:
                interpolate_met(met, self.source, key, **self.interp_kwargs)

            self.source["rho_air"] = thermo.rho_d(
                self.source["air_temperature"], self.source.air_pressure
            )

        # STEP 5: Calculate segment-specific properties if they are not already attached
        if "true_airspeed" not in self.source:
            self.source["true_airspeed"] = self.source.segment_true_airspeed(
                u_wind=self.source["u_wind"],
                v_wind=self.source["v_wind"],
                smooth=self.params["smooth_true_airspeed"],
                window_length=self.params["smooth_true_airspeed_window_length"],
            )
        if "segment_length" not in self.source:
            self.source["segment_length"] = self.source.segment_length()
        max_ = np.nanmax(self.source["segment_length"])
        lim_ = self.params["max_seg_length_m"]
        if max_ > 0.9 * lim_:
            warnings.warn(
                "Flight trajectory has segment lengths close to or exceeding the "
                "'max_seg_length_m' parameter. Evolved contrail segments may reach "
                "their end of life artificially early. Either resample the flight "
                "with the 'resample_and_fill' method (recommended), or use a larger "
                "'max_seg_length_m' parameter. Current values: "
                f"max_seg_length_m={lim_}, max_seg_length_on_trajectory={max_}"
            )

        # if either angle is not provided, re-calculate both
        # this is the one case where we will overwrite a flight data key
        if "sin_a" not in self.source or "cos_a" not in self.source:
            self.source["sin_a"], self.source["cos_a"] = self.source.segment_angle()

        # STEP 6: Calculate emissions if requested, or if keys don't exist in Flight
        if self.params["process_emissions"]:
            self._process_emissions()

        # STEP 7: Ensure that flight has the required variables defined as attrs or columns
        self.source.ensure_vars(
            (
                "flight_id",
                "engine_efficiency",
                "fuel_flow",
                "aircraft_mass",
                "nvpm_ei_n",
                "wingspan",
            )
        )

    def _process_emissions(self) -> None:
        """Process flight emissions.

        See :class:`Emissions`.

        We should consider supporting OpenAP (https://github.com/TUDelft-CNS-ATM/openap)
        and alternate performance models in the future.
        """
        logger.debug("Processing flight emissions")

        # currently only one emissions model
        emissions = Emissions(copy_source=False)

        # Run against a list of flights (Fleet)
        if isinstance(self.source, Fleet):
            # Rip the Fleet apart, run BADA on each, then reassemble
            logger.debug("Separately running aircraft performance on each flight in fleet")
            fls = self.source.to_flight_list(copy=False)
            fls = [
                _eval_aircraft_performance(self.params["aircraft_performance"], fl) for fl in fls
            ]

            # In Fleet-mode, always call emissions
            logger.debug("Separately running emissions on each flight in fleet")
            fls = [_eval_emissions(emissions, fl) for fl in fls]

            # Need to be careful here -- broadcast_numeric has already been called once,
            # and so data already contains numeric variables attached before bada_flight.eval
            # called. So we just handpick additional attributes to broadcast
            for fl in fls:
                fl.broadcast_attrs("wingspan")

            # Convert back to fleet
            attrs = self.source.attrs
            attrs.pop("fl_attrs", None)
            attrs.pop("data_keys", None)
            self.source = Fleet.from_seq(fls, broadcast_numeric=False, copy=False, attrs=attrs)

        # Single flight
        else:
            self.source = _eval_aircraft_performance(
                self.params["aircraft_performance"], self.source
            )
            self.source = _eval_emissions(emissions, self.source)

        # Scale nvPM with parameter (fleet / flight)
        factor = self.params["nvpm_ei_n_enhancement_factor"]
        try:
            self.source["nvpm_ei_n"] *= factor
        except KeyError:
            self.source.attrs.update({"nvpm_ei_n": self.source.attrs["nvpm_ei_n"] * factor})

    def _find_initial_contrail_regions(self) -> None:
        """Apply Schmidt-Appleman criteria to determine regions of persistent contrail formation.

        This method:
            - Modifies :attr:`flight` in place by assigning additional columns arising from SAC
            - Creates :attr:`_sac_flight` needed for :meth:`_simulate_wave_vortex_downwash`.
        """
        # calculate the SAC for each waypoint
        sac_model = sac.SAC(
            met=None,  # self.source is already interpolated against met, so we don't need it
            copy_source=False,
        )
        self.source = sac_model.eval(source=self.source)

        # Estimate SAC threshold temperature along initial contrail
        # This variable is not involved in calculating the SAC,
        # but it is required by `contrail_properties.ice_particle_number`.
        # The three variables used below are all computed in sac_model.eval
        self.source["T_critical_sac"] = sac.T_critical_sac(
            self.source["T_sat_liquid"],
            self.source["rh"],
            self.source["G"],
        )

        # create a new Flight only at points where "sac" == 1
        filt = self.source["sac"] == 1.0
        if self.params["filter_sac"]:
            self._sac_flight = self.source.filter(filt)
            logger.debug(
                "Fraction of waypoints satisfying the SAC: %s / %s",
                len(self._sac_flight),
                len(self.source),
            )
            return

        warnings.warn("Manually overriding SAC filter")
        logger.info("Manually overriding SAC filter")
        self._sac_flight = self.source.copy()
        logger.debug(
            "Fraction of waypoints satisfying the SAC: %s / %s",
            np.sum(filt),
            self._sac_flight.size,
        )
        logger.debug("None are filtered out!")

    def _simulate_wake_vortex_downwash(self, met: MetDataset) -> None:
        """Apply wake vortex model to calculate initial contrail geometry.

        The calculation uses a parametric wake vortex transport and decay model to simulate the
        wake vortex phase and obtain the initial contrail width, depth and downward displacement.

        This method assigns additional columns "width" and "depth" to :attr:`_sac_flight`

        Parameters
        ----------
        met : MetDataset
            Meteorology data

        References
        ----------
        - :cite:`holzapfelProbabilisticTwoPhaseWake2003`
        """
        air_temperature = self._sac_flight["air_temperature"]
        u_wind = self._sac_flight["u_wind"]
        v_wind = self._sac_flight["v_wind"]
        air_pressure = self._sac_flight.air_pressure

        # flight parameters
        aircraft_mass = self._sac_flight.get_data_or_attr("aircraft_mass")
        true_airspeed = self._sac_flight["true_airspeed"]

        # In Fleet-mode, wingspan resides on `data`, and in Flight-mode,
        # wingspan resides on `attrs`.
        wingspan = self._sac_flight.get_data_or_attr("wingspan")

        # get the pressure level `dz_m` lower than element pressure
        air_pressure_lower = thermo.p_dz(air_temperature, air_pressure, self.params["dz_m"])
        level_lower = air_pressure_lower / 100

        # get full met grid or flight data interpolated to the pressure level `p_dz`
        air_temperature_lower = interpolate_met(
            met,
            self._sac_flight,
            "air_temperature",
            "air_temperature_lower",
            level=level_lower,
            **self.interp_kwargs,
        )
        u_wind_lower = interpolate_met(
            met,
            self._sac_flight,
            "eastward_wind",
            "u_wind_lower",
            level=level_lower,
            **self.interp_kwargs,
        )
        v_wind_lower = interpolate_met(
            met,
            self._sac_flight,
            "northward_wind",
            "v_wind_lower",
            level=level_lower,
            **self.interp_kwargs,
        )

        # Temperature gradient
        dT_dz = self._sac_flight["dT_dz"] = thermo.T_potential_gradient(
            air_temperature,
            air_pressure,
            air_temperature_lower,
            air_pressure_lower,
            self.params["dz_m"],
        )

        # wind shear
        ds_dz = self._sac_flight["ds_dz"] = wind_shear.wind_shear(
            u_wind, u_wind_lower, v_wind, v_wind_lower, self.params["dz_m"]
        )

        # Initial contrail width, depth and downward displacement
        dz_max = self._sac_flight["dz_max"] = wake_vortex.max_downward_displacement(
            wingspan,
            true_airspeed,
            aircraft_mass,
            air_temperature,
            dT_dz,
            ds_dz,
            air_pressure,
            effective_vertical_resolution=self.params["effective_vertical_resolution"],
            wind_shear_enhancement_exponent=self.params["wind_shear_enhancement_exponent"],
        )

        # derive downwash values and save to data model
        self._sac_flight["width"] = wake_vortex.initial_contrail_width(wingspan, dz_max)
        initial_wake_vortex_depth = self.params["initial_wake_vortex_depth"]
        self._sac_flight["depth"] = wake_vortex.initial_contrail_depth(
            dz_max, initial_wake_vortex_depth
        )
        # Initially, sigma_yz is set to 0
        # See bottom left paragraph p. 552 Schumann 2012 beginning with:
        # >>> "The contrail model starts from initial values ..."
        self._sac_flight["sigma_yz"] = np.zeros_like(dz_max)

    def _find_initial_persistent_contrails(self, met: MetDataset) -> None:
        """Calculate the initial contrail properties after the wake vortex phase.

        Determine points with initial persistent contrails.

        This method first calculates ice water content (``iwc``) and number of ice particles per
        distance (``n_ice_per_m_1``).

        It then tests each contrail waypoint for initial persistence and calculates
        parameters for the first iteration of the time integration.

        Note that the subscript "_1" represents the conditions after the wake vortex phase.

        Parameters
        ----------
        met : MetDataset
            Meteorology data
        """

        # met parameters along Flight path
        air_pressure = self._sac_flight.air_pressure
        air_temperature = self._sac_flight["air_temperature"]
        specific_humidity = self._sac_flight["specific_humidity"]
        T_critical_sac = self._sac_flight["T_critical_sac"]

        # Flight performance parameters
        fuel_dist = (
            self._sac_flight.get_data_or_attr("fuel_flow") / self._sac_flight["true_airspeed"]
        )

        nvpm_ei_n = self._sac_flight.get_data_or_attr("nvpm_ei_n")
        ei_h2o = self._sac_flight.fuel.ei_h2o

        # get initial contrail parameters from wake vortex simulation
        width = self._sac_flight["width"]
        depth = self._sac_flight["depth"]

        # initial contrail altitude set to 0.5 * depth
        contrail_1 = GeoVectorDataset(
            longitude=self._sac_flight["longitude"],
            latitude=self._sac_flight["latitude"],
            altitude=self._sac_flight.altitude - 0.5 * depth,
            time=self._sac_flight["time"],
            copy=False,
        )

        # get met post wake vortex along initial contrail
        air_temperature_1 = interpolate_met(
            met,
            contrail_1,
            "air_temperature",
            **self.interp_kwargs,
        )
        interpolate_met(
            met,
            contrail_1,
            "specific_humidity",
            **self.interp_kwargs,
        )

        if self.params["humidity_scaling"] is not None:
            self.params["humidity_scaling"].eval(contrail_1, copy_source=False)
        else:
            contrail_1["air_pressure"] = contrail_1.air_pressure
            contrail_1["rhi"] = thermo.rhi(
                contrail_1["specific_humidity"], air_temperature_1, contrail_1["air_pressure"]
            )

        air_pressure_1 = contrail_1["air_pressure"]
        specific_humidity_1 = contrail_1["specific_humidity"]
        rhi_1 = contrail_1["rhi"]
        level_1 = contrail_1.level
        altitude_1 = contrail_1.altitude

        # calculate thermo properties post wake vortex
        q_sat_1 = thermo.q_sat_ice(air_temperature_1, air_pressure_1)
        rho_air_1 = thermo.rho_d(air_temperature_1, air_pressure_1)

        # Initialize initial contrail properties
        iwc = contrail_properties.initial_iwc(
            air_temperature, specific_humidity, air_pressure, fuel_dist, width, depth, ei_h2o
        )
        iwc_ad = contrail_properties.iwc_adiabatic_heating(
            air_temperature, air_pressure, air_pressure_1
        )
        iwc_1 = contrail_properties.iwc_post_wake_vortex(iwc, iwc_ad)
        n_ice_per_m_1 = contrail_properties.ice_particle_number(
            nvpm_ei_n=nvpm_ei_n,
            fuel_dist=fuel_dist,
            iwc=iwc,
            iwc_1=iwc_1,
            air_temperature=air_temperature,
            T_crit_sac=T_critical_sac,
            min_ice_particle_number_nvpm_ei_n=self.params["min_ice_particle_number_nvpm_ei_n"],
        )

        # Check for persistent initial_contrails
        persistent_1 = contrail_properties.initial_persistent(iwc_1, rhi_1)

        self._sac_flight["altitude_1"] = altitude_1
        self._sac_flight["level_1"] = level_1
        self._sac_flight["air_temperature_1"] = air_temperature_1
        self._sac_flight["specific_humidity_1"] = specific_humidity_1
        self._sac_flight["q_sat_1"] = q_sat_1
        self._sac_flight["air_pressure_1"] = air_pressure_1
        self._sac_flight["rho_air_1"] = rho_air_1
        self._sac_flight["rhi_1"] = rhi_1
        self._sac_flight["iwc_1"] = iwc_1
        self._sac_flight["n_ice_per_m_1"] = n_ice_per_m_1
        self._sac_flight["persistent_1"] = persistent_1

        # Create new Flight only at persistent points
        if self.params["filter_initially_persistent"]:
            self._downwash_flight = self._sac_flight.filter(persistent_1.astype(bool))
            logger.debug(
                "Fraction of waypoints with initially persistent contrails: %s / %s",
                len(self._downwash_flight),
                len(self._sac_flight),
            )

        else:
            warnings.warn("Manually overriding initially persistent filter")
            logger.info("Manually overriding initially persistent filter")
            self._downwash_flight = self._sac_flight.copy()
            logger.debug(
                "Fraction of waypoints with initially persistent contrails: %s / %s",
                persistent_1.sum(),
                persistent_1.size,
            )
            logger.debug("None are filtered out!")

    def _simulate_contrail_evolution(self) -> None:
        """Simulate contrail evolution."""
        # Calculate all properties for "downwash_contrail" which is
        # a contrail representation of the waypoints of the downwash flight.
        # The downwash_contrail has already been filtered for initial persistent waypoints.
        self._downwash_contrail = self._create_downwash_contrail()
        buffers = {
            f"{coord}_buffer": self.params[f"met_{coord}_buffer"]
            for coord in ["longitude", "latitude", "level"]
        }
        logger.debug("Downselect met for start of Cocip evolution")
        met = self._downwash_contrail.downselect_met(self.met, **buffers, copy=False)
        rad = self._downwash_contrail.downselect_met(self.rad, **buffers, copy=False)

        calc_continuous(self._downwash_contrail)
        calc_timestep_geometry(self._downwash_contrail)
        calc_timestep_meteorology(self._downwash_contrail, met, self.params, self.interp_kwargs)
        calc_shortwave_radiation(rad, self._downwash_contrail, **self.interp_kwargs)
        calc_outgoing_longwave_radiation(rad, self._downwash_contrail, **self.interp_kwargs)
        calc_contrail_properties(
            self._downwash_contrail,
            self.params["effective_vertical_resolution"],
            self.params["wind_shear_enhancement_exponent"],
            self.params["sedimentation_impact_factor"],
            self.params["radiative_heating_effects"],
        )

        # Intersect with rad dataset
        calc_radiative_properties(self._downwash_contrail, self.params)

        # Complete iteration at time_idx - 2
        for time_idx, time_end in enumerate(self.timesteps[:-1]):
            logger.debug("Start time integration step %s ending at time %s", time_idx, time_end)

            # get the last evolution step of contrail waypoints, if it exists
            if self.contrail_list:
                latest_contrail = self.contrail_list[-1]
            else:
                latest_contrail = GeoVectorDataset()

            # load new contrail segments from downwash_flight
            contrail_2_segments = self._get_contrail_2_segments(time_idx)
            if contrail_2_segments:
                logger.debug(
                    "Discover %s new contrail waypoints formed by downwash_flight waypoints.",
                    contrail_2_segments.size,
                )
                logger.debug("Previously persistent contrail size: %s", latest_contrail.size)

                # Append new waypoints to latest contrail, or set as first contrail waypoints
                latest_contrail = latest_contrail + contrail_2_segments

                # CRITICAL: When running in "Fleet" mode, the latest_contrail is no longer
                # sorted by flight_id. Fixing that below.
                if isinstance(self.source, Fleet):
                    latest_contrail = latest_contrail.sort(["flight_id", "time"])

            # Check for an empty contrail
            if not latest_contrail:
                logger.debug("Empty latest_contrail at timestep %s", time_end)
                if np.all(time_end > self._downwash_contrail["time"]):
                    logger.debug("No remaining downwash_contrail waypoints. Break.")
                    break
                continue

            # Update met, rad slices as needed
            # We need to both interpolate latest_contrail, as well as the "contrail_2"
            # created by calc_timestep_contrail_evolution. This "contrail_2" object
            # has constant time at "time_end", hence the buffer we apply below.
            # After the downwash_contrails is all used up, these updates are intended
            # to happen once each hour
            buffers["time_buffer"] = (
                np.timedelta64(0, "ns"),
                time_end - latest_contrail["time"].max(),
            )
            if time_end > met.variables["time"].values[-1]:
                logger.debug("Downselect met at time_end %s within Cocip evolution", time_end)
                met = latest_contrail.downselect_met(self.met, **buffers, copy=False)
            if time_end > rad.variables["time"].values[-1]:
                logger.debug("Downselect rad at time_end %s within Cocip evolution", time_end)
                rad = latest_contrail.downselect_met(self.rad, **buffers, copy=False)

            # Recalculate latest_contrail with new values
            # NOTE: We are doing a substantial amount of redundant computation here
            # At waypoints for which the continuity hasn't changed, there is nothing
            # new going on, and so we are overwriting variables in latest_contrail
            # with the same values
            # NOTE: Both latest_contrail and contrail_2_segments contain all
            # 54 variables. The only difference is the change in continuity.
            # The change in continuity impacts:
            # - sin_a, cos_a, segment_length (in calc_timestep_geometry)
            # - dsn_dz (in calc_timestep_meteorology, then enhanced in contrail_properties)
            # And there is huge room to optimize this
            calc_continuous(latest_contrail)
            calc_timestep_geometry(latest_contrail)
            calc_timestep_meteorology(latest_contrail, met, self.params, self.interp_kwargs)
            calc_contrail_properties(
                latest_contrail,
                self.params["effective_vertical_resolution"],
                self.params["wind_shear_enhancement_exponent"],
                self.params["sedimentation_impact_factor"],
                self.params["radiative_heating_effects"],
            )

            final_contrail = calc_timestep_contrail_evolution(
                met=met,
                rad=rad,
                contrail_1=latest_contrail,
                time_2=time_end,
                params=self.params,
                **self.interp_kwargs,
            )
            self.contrail_list.append(final_contrail)

    def _create_downwash_contrail(self) -> GeoVectorDataset:
        """Get Contrail representation of downwash flight."""

        downwash_contrail_data = {
            "waypoint": self._downwash_flight["waypoint"],
            "flight_id": self._downwash_flight["flight_id"],
            "time": self._downwash_flight["time"],
            "longitude": self._downwash_flight["longitude"],
            "latitude": self._downwash_flight["latitude"],
            # intentionally specify altitude and level to avoid pressure level calculations
            "altitude": self._downwash_flight["altitude_1"],
            "level": self._downwash_flight["level_1"],
            "air_pressure": self._downwash_flight["air_pressure_1"],
            "width": self._downwash_flight["width"],
            "depth": self._downwash_flight["depth"],
            "sigma_yz": self._downwash_flight["sigma_yz"],
            "air_temperature": self._downwash_flight["air_temperature_1"],
            "specific_humidity": self._downwash_flight["specific_humidity_1"],
            "q_sat": self._downwash_flight["q_sat_1"],
            "rho_air": self._downwash_flight["rho_air_1"],
            "rhi": self._downwash_flight["rhi_1"],
            "iwc": self._downwash_flight["iwc_1"],
            "n_ice_per_m": self._downwash_flight["n_ice_per_m_1"],
            "persistent": self._downwash_flight["persistent_1"],
        }

        contrail = GeoVectorDataset(downwash_contrail_data, copy=True)
        contrail["formation_time"] = contrail["time"].copy()
        contrail["age"] = contrail["formation_time"] - contrail["time"]

        # Heating rate, differential heating rate and
        # cumulative heat energy absorbed by the contrail
        if self.params["radiative_heating_effects"]:
            contrail["heat_rate"] = np.zeros_like(contrail["n_ice_per_m"])
            contrail["d_heat_rate"] = np.zeros_like(contrail["n_ice_per_m"])

            # Increase in temperature averaged over the contrail plume
            contrail["cumul_heat"] = np.zeros_like(contrail["n_ice_per_m"])

            # Temperature difference between the upper and lower half of the contrail plume
            contrail["cumul_differential_heat"] = np.zeros_like(contrail["n_ice_per_m"])

        # Initially set energy forcing to 0 because the contrail just formed (age = 0)
        contrail["ef"] = np.zeros_like(contrail["n_ice_per_m"])

        if not self.params["filter_sac"]:
            contrail["sac"] = self._downwash_flight["sac"]
        if not self.params["filter_initially_persistent"]:
            contrail["initially_persistent"] = self._downwash_flight["persistent_1"]
        if self.params["persistent_buffer"] is not None:
            contrail["end_of_life"] = np.full(contrail.size, np.datetime64("NaT", "ns"))

        return contrail

    def _get_contrail_2_segments(self, time_idx: int) -> GeoVectorDataset:
        """Get batch of newly formed contrail segments."""
        time = self._downwash_contrail["time"]
        t_cur = self.timesteps[time_idx]
        if time_idx == 0:
            filt = time < t_cur
        else:
            t_prev = self.timesteps[time_idx - 1]
            filt = (time < t_cur) & (time >= t_prev)

        return self._downwash_contrail.filter(filt)

    @overrides
    def _cleanup_indices(self) -> None:
        if self.params["interpolation_use_indices"]:
            for contrail in self.contrail_list:
                contrail._invalidate_indices()
            self.source._invalidate_indices()
            self._sac_flight._invalidate_indices()
            self._downwash_flight._invalidate_indices()
            self._downwash_contrail._invalidate_indices()

    def _bundle_results(self) -> None:
        # ---
        # Create contrail dataframe (self.contrail)
        # ---
        dfs = [contrail.dataframe for contrail in self.contrail_list]
        dfs = [df.assign(timestep=t_idx) for t_idx, df in enumerate(dfs)]
        self.contrail = pd.concat(dfs)

        # add age in hours to the contrail waypoint outputs
        self.contrail["age_hours"] = self.contrail["age"] / np.timedelta64(1, "h")

        if self.params["verbose_outputs"]:
            # Compute dt_integration -- logic is somewhat complicated, but
            # we're simply addressing that the first dt_integration
            # is different from the rest

            # We call reset_index twice. The first call introduces an `index`
            # column, and the second introduces a `level_0` column. This `level_0`
            # is a RangeIndex, which we use in the `groupby` to identify the
            # index of the first evolution step at each waypoint.
            # The `level_0` is used to insert back into the `seq_index` dataframe,
            # then it is dropped in replace of the original `index`.
            seq_index = self.contrail.reset_index().reset_index()
            cols = ["formation_time", "time", "level_0"]
            first_form_time = seq_index.groupby("waypoint")[cols].first()
            first_dt = first_form_time["time"] - first_form_time["formation_time"]
            first_dt.index = first_form_time["level_0"]

            seq_index = seq_index.set_index("level_0")
            seq_index["dt_integration"] = first_dt
            seq_index["dt_integration"].fillna(self.params["dt_integration"], inplace=True)

            self.contrail = seq_index.set_index("index")

        # ---
        # Create contrail xr.Dataset (self.contrail_dataset)
        # ---
        if self.params["verbose_outputs"]:
            if isinstance(self.source, Fleet):
                self.contrail_dataset = xr.Dataset.from_dataframe(
                    self.contrail.set_index(["flight_id", "timestep", "waypoint"])
                )
            else:
                self.contrail_dataset = xr.Dataset.from_dataframe(
                    self.contrail.set_index(["timestep", "waypoint"])
                )

        # ---
        # Create output Flight / Fleet (self.source)
        # ---

        col_idx = ["flight_id", "waypoint"] if isinstance(self.source, Fleet) else ["waypoint"]
        del self.source["_met_intersection"]

        # Attach intermediate calculations from `sac_flight` and `downwash_flight` to flight
        sac_cols = [
            "width",
            "depth",
            "rhi_1",
            "air_temperature_1",
            "specific_humidity_1",
            "altitude_1",
            "persistent_1",
        ]

        # add additional columns
        if self.params["verbose_outputs"]:
            sac_cols += ["dT_dz", "ds_dz", "dz_max"]

        downwash_cols = ["rho_air_1", "iwc_1", "n_ice_per_m_1"]
        df = pd.concat(
            [
                self.source.dataframe.set_index(col_idx),
                self._sac_flight.dataframe.set_index(col_idx)[sac_cols],
                self._downwash_flight.dataframe.set_index(col_idx)[downwash_cols],
            ],
            axis=1,
        )

        # Aggregate contrail data back to flight
        grouped = self.contrail.groupby(col_idx)

        # Perform all aggregations
        agg_dict = {"ef": ["sum"], "age": ["max"]}
        rad_keys = ["sdr", "rsr", "olr", "rf_sw", "rf_lw", "rf_net"]
        for key in rad_keys:
            if self.params["verbose_outputs"]:
                agg_dict[key] = ["mean", "min", "max"]
            else:
                agg_dict[key] = ["mean"]

        aggregated = grouped.agg(agg_dict)
        aggregated.columns = [f"{k1}_{k2}" for k1, k2 in aggregated.columns]
        aggregated = aggregated.rename(columns={"ef_sum": "ef", "age_max": "contrail_age"})

        # Join the two
        df = df.join(aggregated)

        # Fill missing values for ef and contrail_age per conventions
        # Mean, max, and min radiative values are *not* filled with 0
        df["ef"] = df["ef"].fillna(0)
        df["contrail_age"] = df["contrail_age"].fillna(np.timedelta64(0, "ns"))

        # cocip flag for each waypoint
        # -1 if negative EF, 0 if no EF, 1 if positive EF,
        # or NaN for outside of domain of flight waypoints that don't persist
        df["cocip"] = np.sign(df["ef"])

        # reset the index
        df = df.reset_index()

        # create new Flight / Fleet output from dataframe with flight attrs

        self.source = type(self.source)(df, attrs=self.source.attrs, copy=False)
        logger.debug("Total number of waypoints with nonzero EF: %s", df["cocip"].ne(0).sum())

    def _fill_empty_flight_results(self, return_list_flight: bool) -> Flight | list[Flight]:
        """Fill empty results into flight / fleet and return.

        This method attaches an all nan array to each of the variables:
            - sdr
            - rsr
            - olr
            - rf_sw
            - rf_lw
            - rf_net

        This method also attaches zeros (for trajectory points contained within met grid)
        or nans (for trajectory points outside of the met grid) to the following variables.
            - ef
            - cocip
            - contrail_age
            - persistent_1

        Parameters
        ----------
        return_list_flight : bool
            If True, a list of :class:`Flight` is returned. In this case, :attr:`source`
            is assumed to be a :class:`Fleet`.

        Returns
        -------
        Flight or list[Flight]
            Flight or list of Flight objects with empty variables.
        """

        empty = np.full(shape=len(self.source), fill_value=np.nan)

        self.source["sdr"] = empty.copy()
        self.source["rsr"] = empty.copy()
        self.source["olr"] = empty.copy()
        self.source["rf_sw"] = empty.copy()
        self.source["rf_lw"] = empty.copy()
        self.source["rf_net"] = empty.copy()

        intersection = self.source.data.pop("_met_intersection")
        zeros_and_nans = np.where(intersection, 0.0, np.nan)
        self.source["ef"] = zeros_and_nans.copy()
        self.source["persistent_1"] = zeros_and_nans.copy()
        self.source["cocip"] = np.sign(zeros_and_nans)
        self.source["contrail_age"] = zeros_and_nans.astype("timedelta64[ns]")

        if return_list_flight:
            return self.source.to_flight_list()  # type: ignore[attr-defined]

        return self.source


# ----------------------------------------
# Functions used in Cocip and CocipGrid
# ----------------------------------------


def process_met_datasets(
    met: MetDataset,
    rad: MetDataset,
    shift_radiation_time: np.timedelta64 | None = None,
) -> tuple[MetDataset, MetDataset]:
    """Process and verify ERA5 data for :class:`Cocip` and :class:`CocipGrid`.

    The implementation uses :class:`Cocip` for the source of truth in determining
    which met variables are required.

    .. versionchanged:: 0.25.0

        This method is called in the :class:`CocipGrid` constructor regardless
        of the ``process_met`` parameter. The same approach is also taken
        in :class:`Cocip` in version 0.27.0.

    Parameters
    ----------
    met : MetDataset
        Met pressure-level data
    rad : MetDataset
        Rad single-level data
    shift_radiation_time : np.timedelta64 | None
        Shift the time dimension of radiation data to account for accumulated values.
        If not specified, the default value from :class:`CocipGridParams` will be used.

    Returns
    -------
    met : MetDataset
        Met data with "tau_cirrus" variable attached.
    rad : MetDataset
        Rad data with time shifted to account for accumulated values.
    """
    # Check for remnants of previous scaling.
    if "_pycontrails_modified" in met["specific_humidity"].attrs:
        raise ValueError(
            "Specific humidity enhancement of the raw specific humidity values in "
            "the underlying met data is deprecated."
        )
    if "tau_cirrus" not in met.data:
        met.ensure_vars(Cocip.met_variables)
        met.data["tau_cirrus"] = tau_cirrus.tau_cirrus(met)
    else:
        met.ensure_vars(Cocip.processed_met_variables)

    # Deal with rad: check shift_radiation_time
    rad.ensure_vars(Cocip.rad_variables)
    if "_pycontrails_modified" in rad["time"].attrs:
        existing_shift = rad["time"].attrs["shift_radiation_time"]  # this is a string
        if pd.Timedelta(existing_shift) != shift_radiation_time:  # compare timedeltas
            raise ValueError(
                "The time coordinate in MetDataset 'rad' has already been "
                f"scaled by a CoCiP model with 'shift_radiation_time={existing_shift}'. "
            )
    else:
        shift_radiation_time = shift_radiation_time or Cocip.default_params.shift_radiation_time
        rad = _process_rad(rad, shift_radiation_time)

    return met, rad


def _process_rad(rad: MetDataset, shift_radiation_time: np.timedelta64) -> MetDataset:
    """Process radiation specific variables for model.

    These variables are used to calculate the reflected solar radiation (RSR),
    and outgoing longwave radiation (OLR).

    The time stamp is adjusted by ``shift_radiation_time`` to account for
    accumulated values being averaged over the time period.

    Parameters
    ----------
    rad : MetDataset
        Rad single-level data
    shift_radiation_time : np.timedelta64
        Shift the time dimension of radiation data to account for accumulated values.

    Returns
    -------
    MetDataset
        Rad data with time shifted.

    Notes
    -----
    - https://www.ecmwf.int/sites/default/files/elibrary/2015/18490-radiation-quantities-ecmwf-model-and-mars.pdf
    - https://confluence.ecmwf.int/pages/viewpage.action?pageId=155337784
    """  # noqa: E501
    logger.debug(f"Shifting radiation time dimension by {shift_radiation_time}")
    if not np.all(np.diff(rad.data["time"]) / 2 == -shift_radiation_time):
        warnings.warn(
            f"Shifting radiation time dimension by unexpected interval {shift_radiation_time}. "
            "If working with ERA5 ensemble members, set "
            "`shift_radiation_time=-np.timedelta64(90, 'm')`. Otherwise, the "
            "expected shift is half the time difference consecutive time steps."
        )

    rad.data = rad.data.assign_coords({"time": rad.data["time"] + shift_radiation_time})
    msg = "Time coordinates adjusted to account for accumulation averaging"
    rad.data["time"].attrs["_pycontrails_modified"] = msg
    rad.data["time"].attrs["shift_radiation_time"] = str(shift_radiation_time)

    return rad


def _eval_aircraft_performance(
    aircraft_performance: AircraftPerformance | None, flight: Flight
) -> Flight:
    """Evaluate the :class:`AircraftPerformance` model.

    Parameters
    ----------
    aircraft_performance : AircraftPerformance | None
        Input aircraft performance model
    flight : Flight
        Flight to evaluate

    Returns
    -------
    Flight
        Output from aircraft performance model

    Raises
    ------
    ValueError
        If ``aircraft_performance`` is None
    """

    aircraft_performance_outputs = "wingspan", "engine_efficiency", "fuel_flow", "aircraft_mass"
    if flight.ensure_vars(aircraft_performance_outputs, False):
        return flight

    if aircraft_performance is None:
        raise ValueError(
            "An AircraftPerformance model parameter is required if the flight "
            f"does not contain the following variables: {aircraft_performance_outputs}"
        )

    return aircraft_performance.eval(source=flight, copy_source=False)


def _eval_emissions(emissions: Emissions, flight: Flight) -> Flight:
    """Evaluate the :class:`Emissions` model.

    Parameters
    ----------
    emissions : Emissions
        Emissions model
    flight : Flight
        Flight to evaluate

    Returns
    -------
    Flight
        Output from emissions model
    """

    emissions_outputs = "nvpm_ei_n"
    if flight.ensure_vars(emissions_outputs, False):
        return flight
    return emissions.eval(source=flight, copy_source=False)


def calc_continuous(contrail: GeoVectorDataset) -> None:
    """Calculate the continuous segments of this timestep.

    Mutates parameter ``contrail`` in place by setting or updating the
    "continuous" variable.

    Parameters
    ----------
    contrail : GeoVectorDataset
        GeoVectorDataset instance onto which "continuous" is set.
    """
    if not contrail:
        raise ValueError("Cannot calculate continuous on an empty contrail")
    same_flight = contrail["flight_id"][:-1] == contrail["flight_id"][1:]
    consecutive_waypoint = np.diff(contrail["waypoint"]) == 1
    continuous = np.empty(contrail.size, dtype=bool)
    continuous[:-1] = same_flight & consecutive_waypoint
    continuous[-1] = False  # This fails if contrail is empty
    contrail.update(continuous=continuous)  # overwrite continuous


def calc_timestep_geometry(contrail: GeoVectorDataset) -> None:
    """Calculate contrail segment-specific properties.

    Mutates parameter ``contrail`` in place by setting or updating the variables
    - "sin_a"
    - "cos_a"
    - "segment_length"

    Any nan values in these variables are set to 0. This ensures any segment-based property
    derived from the segment geometry does not contribute to contrail energy forcing.

    See Also
    --------
    - :func:`wind_shear.wind_shear_normal` to see how "sin_a" and "cos_a"
    are used to compute wind shear terms.
    - :func:`calc_timestep_contrail_evolution` to see how "segment_length" is used.

    Parameters
    ----------
    contrail : GeoVectorDataset
        GeoVectorDataset instance onto which variables are set.
    """
    # get contrail waypoints
    longitude = contrail["longitude"]
    latitude = contrail["latitude"]
    altitude = contrail.altitude
    continuous = contrail["continuous"]

    # calculate segment properties
    segment_length = geo.segment_length(longitude, latitude, altitude)
    sin_a, cos_a = geo.segment_angle(longitude, latitude)

    # NOTE: Set all nan and discontinuous values to 0. With our current implementation, this
    # ensures degenerate or discontinuous segments do not contribute to contrail impact.
    # At the same time, this prevents nan values from propagating through evolving contrails.
    #
    # If we want to change this, take note of the following consequences.
    # nan values flow from one variable to another as follows:
    # segment_length
    # seg_ratio
    # sigma_yz
    # area_eff_2
    # iwc, n_ice_per_m, r_vol_ice
    # terminal_fall_speed
    # level                 [level advection]
    # u_wind, v_wind        [interp]
    # longitude, latitude   [advection]
    # Finally, at the next evolution step, the previous waypoint will accrue
    # a nan value after segment_length is recalculated

    segment_length[~continuous] = 0
    sin_a[~continuous] = 0
    cos_a[~continuous] = 0

    np.nan_to_num(segment_length, copy=False)
    np.nan_to_num(sin_a, copy=False)
    np.nan_to_num(cos_a, copy=False)

    # override values on model
    contrail.update(segment_length=segment_length)
    contrail.update(sin_a=sin_a)
    contrail.update(cos_a=cos_a)


def calc_timestep_meteorology(
    contrail: GeoVectorDataset,
    met: MetDataset,
    params: dict[str, Any],
    interp_kwargs: dict[str, Any],
) -> None:
    """Get and store meteorology parameters.

    Mutates parameter ``contrail`` in place by setting or updating the variables
    - "sin_a"
    - "cos_a"
    - "segment_length"

    Parameters
    ----------
    contrail : GeoVectorDataset
        GeoVectorDataset object onto which meterology variables are attached.
    met : MetDataset
        MetDataset with meteorology data variables.
    params : dict[str, Any]
        Cocip model ``params``.
    interp_kwargs : dict[str, Any]
        Cocip model ``interp_kwargs``.
    """
    # get contrail geometry
    sin_a = contrail["sin_a"]
    cos_a = contrail["cos_a"]

    # get standard met parameters for timestep
    air_pressure = contrail.air_pressure
    air_temperature = contrail["air_temperature"]
    u_wind = interpolate_met(met, contrail, "eastward_wind", "u_wind", **interp_kwargs)
    v_wind = interpolate_met(met, contrail, "northward_wind", "v_wind", **interp_kwargs)
    interpolate_met(
        met,
        contrail,
        "lagrangian_tendency_of_air_pressure",
        "vertical_velocity",
        **interp_kwargs,
    )
    interpolate_met(met, contrail, "tau_cirrus", **interp_kwargs)

    # get the pressure level `dz_m` lower than element pressure
    air_pressure_lower = thermo.p_dz(air_temperature, air_pressure, params["dz_m"])

    # get meteorology at contrail waypoints interpolated at the pressure level `air_pressure_lower`
    level_lower = air_pressure_lower / 100  # Pa -> hPa

    # if met is already interpolated, this will automatically skip interpolation
    air_temperature_lower = interpolate_met(
        met,
        contrail,
        "air_temperature",
        "air_temperature_lower",
        level=level_lower,
        **interp_kwargs,
    )
    u_wind_lower = interpolate_met(
        met,
        contrail,
        "eastward_wind",
        "u_wind_lower",
        level=level_lower,
        **interp_kwargs,
    )
    v_wind_lower = interpolate_met(
        met,
        contrail,
        "northward_wind",
        "v_wind_lower",
        level=level_lower,
        **interp_kwargs,
    )

    # Temperature gradient
    dT_dz = thermo.T_potential_gradient(
        air_temperature,
        air_pressure,
        air_temperature_lower,
        air_pressure_lower,
        params["dz_m"],
    )

    # wind shear
    ds_dz = wind_shear.wind_shear(u_wind, u_wind_lower, v_wind, v_wind_lower, params["dz_m"])

    # wind shear normal
    dsn_dz = wind_shear.wind_shear_normal(
        u_wind_top=u_wind,
        u_wind_btm=u_wind_lower,
        v_wind_top=v_wind,
        v_wind_btm=v_wind_lower,
        cos_a=cos_a,
        sin_a=sin_a,
        dz=params["dz_m"],
    )

    # store values on contrail model
    contrail.update(dT_dz=dT_dz)
    contrail.update(ds_dz=ds_dz)
    contrail.update(dsn_dz=dsn_dz)


def calc_shortwave_radiation(
    rad: MetDataset,
    vector: GeoVectorDataset,
    **kwargs: Any,
) -> None:
    """Calculate shortwave radiation variables.

    Calculates theoretical incident (``sdr``) and
    reflected shortwave radiation (``rsr``) from the radiation data provided.

    Mutates input ``vector`` to include ``"sdr"`` and ``"rsr"``
    keys with calculated SDR, RSR values, respectively.

    Parameters
    ----------
    rad : MetDataset
       Radiation data
    vector : GeoVectorDataset
        Flight or GeoVectorDataset instance
    **kwargs : Any
        Interpolation keyword arguments

    Raises
    ------
    ValueError
        If ``rad`` does not contain ``"toa_upward_shortwave_flux"`` or ``"top_net_solar_radiation"``
        variable.

    Notes
    -----
    In accordance with the original CoCiP Fortran algorithm,
    the SDR is set to 0 when the cosine of the solar zenith angle is below 0.01.

    See Also
    --------
    :func:`geo.solar_direct_radiation`
    """

    if "sdr" in vector and "rsr" in vector:
        return None

    # calculate instantaneous theoretical solar direct radiation based on geo position and time
    longitude = vector["longitude"]
    latitude = vector["latitude"]
    time = vector["time"]
    vector["sdr"] = geo.solar_direct_radiation(longitude, latitude, time, threshold_cos_sza=0.01)

    # GFS contains RSR (toa_upward_shortwave_flux) variable directly
    if "toa_upward_shortwave_flux" in rad:
        interpolate_met(rad, vector, "toa_upward_shortwave_flux", "rsr", **kwargs)

    # ECMWF contains "top_net_solar_radiation" which is SDR - RSR
    elif "top_net_solar_radiation" in rad:
        interpolate_met(rad, vector, "top_net_solar_radiation", **kwargs)
        vector["rsr"] = np.maximum(vector["sdr"] - vector["top_net_solar_radiation"], 0)

    else:
        raise ValueError(
            "'rad' data must contain either 'toa_upward_shortwave_flux' or "
            "'top_net_solar_radiation' (ECMWF) variable."
        )


def calc_outgoing_longwave_radiation(
    rad: MetDataset,
    vector: GeoVectorDataset,
    **kwargs: Any,
) -> None:
    """Calculate outgoing longwave radiation (``olr``) from the radiation data provided.

    Mutates input ``vector`` to include ``"olr"`` key with calculated OLR values.

    Parameters
    ----------
    rad : MetDataset
       Radiation data
    vector : GeoVectorDataset
        Flight or GeoVectorDataset instance
    **kwargs : Any
        Interpolation keyword arguments

    Raises
    ------
    ValueError
        If ``rad`` does not contain a ``"toa_upward_longwave_flux"``
        or ``"top_net_thermal_radiation"`` variable.
    """

    if "olr" in vector:
        return None

    # GFS contains OLR (toa_upward_longwave_flux) variable directly
    if "toa_upward_longwave_flux" in rad:
        interpolate_met(rad, vector, "toa_upward_longwave_flux", "olr", **kwargs)
        return

    # ECMWF contains "top_net_thermal_radiation" which is -1 * OLR
    if "top_net_thermal_radiation" in rad:
        interpolate_met(rad, vector, "top_net_thermal_radiation", **kwargs)
        vector["olr"] = np.maximum(-vector["top_net_thermal_radiation"], 0.0)
        return

    raise ValueError(
        "rad data must contain either 'toa_upward_longwave_flux' "
        "or 'top_net_thermal_radiation' (ECMWF) variable."
    )


def calc_radiative_properties(contrail: GeoVectorDataset, params: dict[str, Any]) -> None:
    """Calculate radiative properties for contrail.

    This function is used by both :class:`Cocip` and :class`CocipGrid`.

    Mutates original `contrail` parameter with additional keys:
        - "rf_sw"
        - "rf_lw"
        - "rf_net"

    Parameters
    ----------
    contrail : GeoVectorDataset
        Grid points already interpolated against met and rad data. In particular,
        the variables
            - "air_temperature"
            - "r_ice_vol"
            - "tau_contrail"
            - "tau_cirrus"
            - "olr"
            - "rsr"
            - "sdr"

        are required on the parameter `contrail`.
    params : dict[str, Any]
        Model parameters
    """
    time = contrail["time"]
    air_temperature = contrail["air_temperature"]

    if params["radiative_heating_effects"]:
        air_temperature += contrail["cumul_heat"]

    r_ice_vol = contrail["r_ice_vol"]
    tau_contrail = contrail["tau_contrail"]
    tau_cirrus_ = contrail["tau_cirrus"]

    # calculate solar constant
    theta_rad = geo.orbital_position(time)
    sd0 = geo.solar_constant(theta_rad)

    # radiation dataset with contrail waypoints at timestep
    sdr = contrail["sdr"]
    rsr = contrail["rsr"]
    olr = contrail["olr"]

    # radiation calculations
    r_vol_um = r_ice_vol * 1e6
    habit_weights = radiative_forcing.habit_weights(
        r_vol_um, params["habit_distributions"], params["radius_threshold_um"]
    )
    rf_lw = radiative_forcing.longwave_radiative_forcing(
        r_vol_um, olr, air_temperature, tau_contrail, tau_cirrus_, habit_weights
    )
    rf_sw = radiative_forcing.shortwave_radiative_forcing(
        r_vol_um, sdr, rsr, sd0, tau_contrail, tau_cirrus_, habit_weights
    )

    # scale RF by enhancement factors
    rf_lw_scaled = rf_lw * params["rf_lw_enhancement_factor"]
    rf_sw_scaled = rf_sw * params["rf_sw_enhancement_factor"]
    rf_net = radiative_forcing.net_radiative_forcing(rf_lw_scaled, rf_sw_scaled)

    # store values on contrail
    contrail["rf_sw"] = rf_sw
    contrail["rf_lw"] = rf_lw
    contrail["rf_net"] = rf_net


def calc_contrail_properties(
    contrail: GeoVectorDataset,
    effective_vertical_resolution: float | np.ndarray,
    wind_shear_enhancement_exponent: float | np.ndarray,
    sedimentation_impact_factor: float | np.ndarray,
    radiative_heating_effects: bool,
) -> None:
    """Calculate geometric and ice-related properties of contrail.

    This function is used by both :class:`Cocip` and :class`CocipGrid`.

    This function modifies parameter `contrail` in place:
        - Mutates contrail data variables:
            - "ds_dz"
            - "dsn_dz"
        - Attaches additional variables:
            - "area_eff"
            - "plume_mass_per_m"
            - "r_ice_vol"
            - "terminal_fall_speed"
            - "diffuse_h"
            - "diffuse_v"
            - "n_ice_per_vol"
            - "tau_contrail"
            - "dn_dt_agg"
            - "dn_dt_turb"

    Parameters
    ----------
    contrail : GeoVectorDataset
        Grid points with many precomputed keys.
    effective_vertical_resolution, wind_shear_enhancement_exponent : float | np.ndarray
        Passed into :func:`wind_shear.wind_shear_enhancement_factor` function.
    sedimentation_impact_factor: float | np.ndarray
        Passed into `contrail_properties.vertical_diffusivity` function.
    radiative_heating_effects: bool
        Include radiative heating effects on contrail cirrus properties.
    """
    time = contrail["time"]
    iwc = contrail["iwc"]
    depth = contrail["depth"]
    width = contrail["width"]
    n_ice_per_m = contrail["n_ice_per_m"]
    t_cirrus_ = contrail["tau_cirrus"]

    # get required meteorology
    air_temperature = contrail["air_temperature"]
    air_pressure = contrail.air_pressure
    rhi = contrail["rhi"]
    rho_air = contrail["rho_air"]
    dT_dz = contrail["dT_dz"]
    ds_dz = contrail["ds_dz"]
    dsn_dz = contrail["dsn_dz"]
    sigma_yz = contrail["sigma_yz"]

    if radiative_heating_effects:
        air_temperature += contrail["cumul_heat"]

    # get required radiation
    theta_rad = geo.orbital_position(time)
    sd0 = geo.solar_constant(theta_rad)
    sdr = contrail["sdr"]
    rsr = contrail["rsr"]
    olr = contrail["olr"]

    # shear enhancements
    shear_enhancement = wind_shear.wind_shear_enhancement_factor(
        contrail_depth=depth,
        effective_vertical_resolution=effective_vertical_resolution,
        wind_shear_enhancement_exponent=wind_shear_enhancement_exponent,
    )
    ds_dz = ds_dz * shear_enhancement
    dsn_dz = dsn_dz * shear_enhancement

    # effective area
    area_eff = contrail_properties.plume_effective_cross_sectional_area(width, depth, sigma_yz)
    depth_eff = contrail_properties.plume_effective_depth(width, area_eff)

    # ice particles
    n_ice_per_vol = contrail_properties.ice_particle_number_per_volume_of_plume(
        n_ice_per_m, area_eff
    )
    n_ice_per_kg_air = contrail_properties.ice_particle_number_per_mass_of_air(
        n_ice_per_vol, rho_air
    )
    plume_mass_per_m = contrail_properties.plume_mass_per_distance(area_eff, rho_air)
    r_ice_vol = contrail_properties.ice_particle_volume_mean_radius(iwc, n_ice_per_kg_air)
    tau_contrail = contrail_properties.contrail_optical_depth(r_ice_vol, n_ice_per_m, width)

    terminal_fall_speed = contrail_properties.ice_particle_terminal_fall_speed(
        air_pressure, air_temperature, r_ice_vol
    )
    diffuse_h = contrail_properties.horizontal_diffusivity(ds_dz, depth)

    if radiative_heating_effects:
        heat_rate = radiative_heating.heating_rate(
            air_temperature=air_temperature,
            rhi=rhi,
            rho_air=rho_air,
            r_ice_vol=r_ice_vol,
            depth_eff=depth_eff,
            tau_contrail=tau_contrail,
            tau_cirrus=t_cirrus_,
            sd0=sd0,
            sdr=sdr,
            rsr=rsr,
            olr=olr,
        )

        cumul_differential_heat = contrail["cumul_differential_heat"]
        d_heat_rate = radiative_heating.differential_heating_rate(
            air_temperature=air_temperature,
            rhi=rhi,
            rho_air=rho_air,
            r_ice_vol=r_ice_vol,
            depth_eff=depth_eff,
            tau_contrail=tau_contrail,
            tau_cirrus=t_cirrus_,
            sd0=sd0,
            sdr=sdr,
            rsr=rsr,
            olr=olr,
        )

        eff_heat_rate = radiative_heating.effective_heating_rate(
            d_heat_rate, cumul_differential_heat, dT_dz, depth
        )
    else:
        eff_heat_rate = None

    diffuse_v = contrail_properties.vertical_diffusivity(
        air_pressure=air_pressure,
        air_temperature=air_temperature,
        dT_dz=dT_dz,
        depth_eff=depth_eff,
        terminal_fall_speed=terminal_fall_speed,
        sedimentation_impact_factor=sedimentation_impact_factor,
        eff_heat_rate=eff_heat_rate,
    )

    dn_dt_agg = contrail_properties.particle_losses_aggregation(
        r_ice_vol, terminal_fall_speed, area_eff
    )
    dn_dt_turb = contrail_properties.particle_losses_turbulence(
        width, depth, depth_eff, diffuse_h, diffuse_v
    )

    # Set properties to model
    contrail.update(ds_dz=ds_dz)
    contrail.update(dsn_dz=dsn_dz)
    contrail.update(area_eff=area_eff)
    contrail.update(plume_mass_per_m=plume_mass_per_m)
    contrail.update(r_ice_vol=r_ice_vol)
    contrail.update(terminal_fall_speed=terminal_fall_speed)
    contrail.update(diffuse_h=diffuse_h)
    contrail.update(diffuse_v=diffuse_v)
    contrail.update(n_ice_per_vol=n_ice_per_vol)
    contrail.update(tau_contrail=tau_contrail)
    contrail.update(dn_dt_agg=dn_dt_agg)
    contrail.update(dn_dt_turb=dn_dt_turb)
    if radiative_heating_effects:
        contrail.update(heat_rate=heat_rate)
        contrail.update(d_heat_rate=d_heat_rate)


def calc_timestep_contrail_evolution(
    met: MetDataset,
    rad: MetDataset,
    contrail_1: GeoVectorDataset,
    time_2: np.datetime64,
    params: dict[str, Any],
    **interp_kwargs: Any,
) -> GeoVectorDataset:
    """Calculate the contrail evolution across timestep (t1 -> t2).

    Note the variable suffix "_1" is used to reference the current time
    and the suffix "_2" is used to refer to the time at the next timestep.

    Parameters
    ----------
    met, rad : MetDataset
       Meteorology and radiation data
    contrail_1 : GeoVectorDataset
        Contrail waypoints at current timestep (1)
    time_2 : np.datetime64
        Time at the end of the evolution step (2)
    params : dict[str, Any]
        Model parameters
    **interp_kwargs : Any
        Interpolation keyword arguments
    """

    # get lat/lon for current timestep (t1)
    longitude_1 = contrail_1["longitude"]
    latitude_1 = contrail_1["latitude"]
    level_1 = contrail_1.level
    time_1 = contrail_1["time"]

    # get contrail_1 geometry
    segment_length_1 = contrail_1["segment_length"]
    width_1 = contrail_1["width"]
    depth_1 = contrail_1["depth"]

    # get required met values for evolution calculations
    q_sat_1 = contrail_1["q_sat"]
    rho_air_1 = contrail_1["rho_air"]
    u_wind_1 = contrail_1["u_wind"]
    v_wind_1 = contrail_1["v_wind"]

    specific_humidity_1 = contrail_1["specific_humidity"]
    vertical_velocity_1 = contrail_1["vertical_velocity"]
    iwc_1 = contrail_1["iwc"]

    # get required contrail_1 properties
    sigma_yz_1 = contrail_1["sigma_yz"]
    dsn_dz_1 = contrail_1["dsn_dz"]
    terminal_fall_speed_1 = contrail_1["terminal_fall_speed"]
    diffuse_h_1 = contrail_1["diffuse_h"]
    diffuse_v_1 = contrail_1["diffuse_v"]
    plume_mass_per_m_1 = contrail_1["plume_mass_per_m"]
    dn_dt_agg_1 = contrail_1["dn_dt_agg"]
    dn_dt_turb_1 = contrail_1["dn_dt_turb"]
    n_ice_per_m_1 = contrail_1["n_ice_per_m"]

    # get contrail_1 radiative properties
    rf_net_1 = contrail_1["rf_net"]

    # initialize new timestep with evolved coordinates
    # assume waypoints are the same to start
    waypoint_2 = contrail_1["waypoint"]
    formation_time_2 = contrail_1["formation_time"]
    time_2_array = np.full_like(time_1, time_2)
    dt = time_2_array - time_1

    # get new contrail location & segment properties after t_step
    longitude_2 = geo.advect_longitude(longitude_1, latitude_1, u_wind_1, dt)
    latitude_2 = geo.advect_latitude(latitude_1, v_wind_1, dt)
    level_2 = geo.advect_level(level_1, vertical_velocity_1, rho_air_1, terminal_fall_speed_1, dt)
    altitude_2 = units.pl_to_m(level_2)

    data = VectorDataDict(
        {
            "waypoint": waypoint_2,
            "flight_id": contrail_1["flight_id"],
            "formation_time": formation_time_2,
            "time": time_2_array,
            "age": time_2_array - formation_time_2,
            "longitude": longitude_2,
            "latitude": latitude_2,
            "altitude": altitude_2,
            "level": level_2,
        }
    )
    contrail_2 = GeoVectorDataset(data, copy=False)

    # Update cumulative radiative heating energy absorbed by the contrail
    # This will always be zero if radiative_heating_effects is not activated in cocip_params
    if params["radiative_heating_effects"]:
        dt_sec = dt / np.timedelta64(1, "s")
        heat_rate_1 = contrail_1["heat_rate"]
        cumul_heat = contrail_1["cumul_heat"]
        cumul_heat += heat_rate_1 * dt_sec
        cumul_heat.clip(max=1.5, out=cumul_heat)  # Constrain additional heat to 1.5 K as precaution
        contrail_2["cumul_heat"] = cumul_heat

        d_heat_rate_1 = contrail_1["d_heat_rate"]
        cumul_differential_heat = contrail_1["cumul_differential_heat"]
        cumul_differential_heat += -d_heat_rate_1 * dt_sec
        contrail_2["cumul_differential_heat"] = cumul_differential_heat

    # Attach a few more artifacts for disabled filtering
    if not params["filter_sac"]:
        contrail_2["sac"] = contrail_1["sac"]
    if not params["filter_initially_persistent"]:
        contrail_2["initially_persistent"] = contrail_1["initially_persistent"]
    if params["persistent_buffer"] is not None:
        contrail_2["end_of_life"] = contrail_1["end_of_life"]

    # calculate initial contrail properties for the next timestep
    calc_continuous(contrail_2)
    calc_timestep_geometry(contrail_2)

    # get next segment lengths
    segment_length_2 = contrail_2["segment_length"]

    # new contrail dimensions
    seg_ratio_2 = contrail_properties.segment_length_ratio(segment_length_1, segment_length_2)
    sigma_yy_2, sigma_zz_2, sigma_yz_2 = contrail_properties.plume_temporal_evolution(
        width_1,
        depth_1,
        sigma_yz_1,
        dsn_dz_1,
        diffuse_h_1,
        diffuse_v_1,
        seg_ratio_2,
        dt,
        max_contrail_depth=params["max_contrail_depth"],
    )
    width_2, depth_2 = contrail_properties.new_contrail_dimensions(sigma_yy_2, sigma_zz_2)

    # store data back in model to support next time step calculations
    contrail_2["width"] = width_2
    contrail_2["depth"] = depth_2
    contrail_2["sigma_yz"] = sigma_yz_2

    # new contrail meteorology parameters
    air_temperature_2 = interpolate_met(met, contrail_2, "air_temperature", **interp_kwargs)

    if params["radiative_heating_effects"]:
        air_temperature_2 += contrail_2["cumul_heat"]

    interpolate_met(met, contrail_2, "specific_humidity", **interp_kwargs)

    if params["humidity_scaling"] is not None:
        params["humidity_scaling"].eval(contrail_2, copy_source=False)
    else:
        contrail_2["air_pressure"] = contrail_2.air_pressure
        contrail_2["rhi"] = thermo.rhi(
            contrail_2["specific_humidity"], air_temperature_2, contrail_2["air_pressure"]
        )

    air_pressure_2 = contrail_2["air_pressure"]
    specific_humidity_2 = contrail_2["specific_humidity"]
    rho_air_2 = thermo.rho_d(air_temperature_2, air_pressure_2)
    q_sat_2 = thermo.q_sat_ice(air_temperature_2, air_pressure_2)

    # store data back in model to support next ``calc_timestep_meteorology``
    contrail_2["rho_air"] = rho_air_2
    contrail_2["q_sat"] = q_sat_2

    # New contrail ice particle mass and number
    area_eff_2 = contrail_properties.new_effective_area_from_sigma(
        sigma_yy_2, sigma_zz_2, sigma_yz_2
    )
    plume_mass_per_m_2 = contrail_properties.plume_mass_per_distance(area_eff_2, rho_air_2)
    iwc_2 = contrail_properties.new_ice_water_content(
        iwc_1,
        specific_humidity_1,
        specific_humidity_2,
        q_sat_1,
        q_sat_2,
        plume_mass_per_m_1,
        plume_mass_per_m_2,
    )
    n_ice_per_m_2 = contrail_properties.new_ice_particle_number(
        n_ice_per_m_1, dn_dt_agg_1, dn_dt_turb_1, seg_ratio_2, dt
    )

    contrail_2["n_ice_per_m"] = n_ice_per_m_2
    contrail_2["iwc"] = iwc_2

    # calculate next timestep meteorology, contrail, and radiative properties
    calc_timestep_meteorology(contrail_2, met, params, interp_kwargs)

    # Intersect with rad dataset
    calc_shortwave_radiation(rad, contrail_2, **interp_kwargs)
    calc_outgoing_longwave_radiation(rad, contrail_2, **interp_kwargs)
    calc_contrail_properties(
        contrail_2,
        params["effective_vertical_resolution"],
        params["wind_shear_enhancement_exponent"],
        params["sedimentation_impact_factor"],
        params["radiative_heating_effects"],
    )
    calc_radiative_properties(contrail_2, params)

    # get properties to measure persistence
    latitude_2 = contrail_2["latitude"]
    altitude_2 = contrail_2.altitude
    age_2 = contrail_2["age"]
    tau_contrail_2 = contrail_2["tau_contrail"]
    n_ice_per_vol_2 = contrail_2["n_ice_per_vol"]

    # calculate next persistent
    persistent_2 = contrail_2["persistent"] = contrail_properties.contrail_persistent(
        latitude=latitude_2,
        altitude=altitude_2,
        segment_length=segment_length_2,
        age=age_2,
        tau_contrail=tau_contrail_2,
        n_ice_per_m3=n_ice_per_vol_2,
        params=params,
    )

    # Get energy forcing by looking forward to the next time step radiative forcing
    rf_net_2 = contrail_2["rf_net"]
    energy_forcing_2 = contrail_properties.energy_forcing(
        rf_net_t1=rf_net_1,
        rf_net_t2=rf_net_2,
        width_t1=width_1,
        width_t2=width_2,
        seg_length_t2=segment_length_2,
        dt=dt,
    )

    # NOTE: Because of our geometry-continuity conventions, any waypoint without
    # continuity automatically has 0 EF. This is because calc_timestep_geometry
    # sets the segment_length entries to 0, thereby eliminating EF.
    # If we change conventions in calc_timestep_geometry, we may want to
    # explicitly set the discontinuous entries here to 0. We do this for age
    # here as well. It's somewhat of a hack, but it ensures nonzero contrail_age
    # is consistent with nonzero EF.
    contrail_2["ef"] = energy_forcing_2
    contrail_2["age"][energy_forcing_2 == 0] = np.timedelta64(0, "ns")

    # filter contrail_2 by persistent waypoints, if any continuous segments are left
    logger.debug(
        "Fraction of waypoints surviving: %s / %s",
        persistent_2.sum(),
        persistent_2.size,
    )

    if (buff := params["persistent_buffer"]) is not None:
        # Here mask gets waypoints that are just now losing persistence
        mask = (~persistent_2) & np.isnat(contrail_2["end_of_life"])
        contrail_2["end_of_life"][mask] = time_2

        # Keep waypoints that are still persistent, which is determined by filt2
        # And waypoints within the persistent buffer, which is determined by filt1
        # So we only drop waypoints that are outside of the persistent buffer
        filt1 = contrail_2["time"] - contrail_2["end_of_life"] < buff
        filt2 = np.isnat(contrail_2["end_of_life"])
        filt = filt1 | filt2
        logger.debug(
            "Fraction of waypoints surviving with buffer %s: %s / %s",
            buff,
            filt.sum(),
            filt.size,
        )
        return contrail_2.filter(filt)

    # filter persistent contrails
    final_contrail = contrail_2.filter(persistent_2)

    # recalculate continuous contrail segments
    # and set EF, age for any newly discontinuous segments to 0
    if final_contrail:
        calc_continuous(final_contrail)
        continuous = final_contrail["continuous"]
        final_contrail["ef"][~continuous] = 0
        final_contrail["age"][~continuous] = np.timedelta64(0, "ns")
    return final_contrail



"""test pycontrails init."""

from __future__ import annotations

import pycontrails


def test_metadata() -> None:
    """Test package data loading."""

    assert pycontrails.__license__ == "Apache-2.0"
    assert pycontrails.__url__



"""Coordinates utilities."""

from __future__ import annotations

import warnings

import numpy as np
import pandas as pd


def slice_domain(
    domain: np.ndarray,
    request: np.ndarray | tuple,
    buffer: tuple[float | np.timedelta64, float | np.timedelta64] = (0.0, 0.0),
) -> slice:
    """Return slice of ``domain`` containing coordinates overlapping ``request``.

    Computes index-based slice (to be used with :meth:`xarray.Dataset.isel` method) as
    opposed to value-based slice.

    Returns ``slice(None, None)`` when ``domain`` has a length <= 2 or the ``request``
    has all ``nan`` values.

    .. versionchanged:: 0.24.1

        The returned slice is the minimum index-based slice that contains the
        requested coordinates. In other words, it's now possible for
        ``domain[sl][0]`` to equal ``min(request)`` where ``sl`` is the output
        of this function. Previously, we were guaranteed that ``domain[sl][0]``
        would be less than ``min(request)``.


    Parameters
    ----------
    domain : np.ndarray
        Full set of domain values
    request : np.ndarray | tuple
        Requested values. Only the min and max values are considered. Either pass in
        a full array-like object or a tuple of ``(min, max)``.
    buffer : tuple[float | np.timedelta64, float | np.timedelta64], optional
        Extend the domain past the requested coordinates by ``buffer[0]`` on the low side
        and ``buffer[1]`` on the high side.
        Units of ``buffer`` must be the same as ``domain``.

    Returns
    -------
    slice
        Slice object for slicing out encompassing, or nearest, domain values

    Raises
    ------
    ValueError
        Raises a ValueError when ``domain`` has all ``nan`` values.

    Examples
    --------
    >>> domain = np.arange(-180, 180, 0.25)

    >>> # Call with request as np.array
    >>> request = np.linspace(-20, 20, 100)
    >>> slice_domain(domain, request)
    slice(640, 801, None)

    >>> # Call with request as tuple
    >>> request = -20, 20
    >>> slice_domain(domain, request)
    slice(640, 801, None)

    >>> # Call with a buffer
    >>> request = -16, 13
    >>> buffer = 4, 7
    >>> slice_domain(domain, request, buffer)
    slice(640, 801, None)

    >>> # Call with request as a single number
    >>> request = -20
    >>> slice_domain(domain, request)
    slice(640, 641, None)

    >>> request = -19.9
    >>> slice_domain(domain, request)
    slice(640, 642, None)

    """
    # if the length of domain coordinates is <= 2, return the whole domain
    if len(domain) <= 2:
        return slice(None, None)

    if buffer == (None, None):
        return slice(None, None)

    # if the request is nan, then there is nothing to slice
    if np.isnan(request).all():
        return slice(None, None)

    # if the whole domain or request is nan, then there is nothing to slice
    if np.isnan(domain).all():
        raise ValueError("Domain is all nan on request")

    # ensure domain is sorted
    zero: float | np.timedelta64 = 0.0
    if pd.api.types.is_datetime64_dtype(domain.dtype):
        zero = np.timedelta64(0)

    if not np.all(np.diff(domain) >= zero):
        raise ValueError("Domain must be sorted in ascending order")

    if np.any(np.asarray(buffer) < zero):
        warnings.warn(
            "Found buffer with negative value. This is unexpected "
            "and will reduce the size of the requested domain instead of "
            "extending it. Both the left and right buffer values should be "
            "nonnegative."
        )

    # get the index of the closest value to request min and max
    # side left returns `i`: domain[i-1] < request <= domain[i]
    # side right returns `i`: domain[i-1] <= request < domain[i]
    idx_min = np.searchsorted(domain, np.nanmin(request) - buffer[0], side="right") - 1
    idx_max = np.searchsorted(domain, np.nanmax(request) + buffer[1], side="left") + 1

    # clip idx_min between [0, len(domain) - 2]
    idx_min = min(len(domain) - 2, max(idx_min, 0))

    # clip idx_max between [2, len(domain)]
    idx_max = min(len(domain), max(idx_max, 2))

    return slice(idx_min, idx_max)


def intersect_domain(
    domain: np.ndarray,
    request: np.ndarray,
) -> np.ndarray:
    """Return boolean mask of ``request`` that are within the bounds of ``domain``.

    Parameters
    ----------
    domain : np.ndarray
        Full set of domain values
    request : np.ndarray
        Full set of requested values

    Returns
    -------
    np.ndarray
        Boolean array of ``request`` values within the bounds of ``domain``

    Raises
    ------
    ValueError
        Raises a ValueError when ``domain`` has all ``nan`` values.
    """
    # if the whole domain or request is nan, then there is nothing to slice
    if np.isnan(domain).all():
        raise ValueError("Domain is all nan on request")

    # if not np.all(np.diff(domain) >= 0):
    #     raise ValueError("Domain must be sorted in ascending order")

    return (request >= np.nanmin(domain)) & (request <= np.nanmax(domain))



"""Interpolation utilities."""

from __future__ import annotations

import dataclasses
import logging
from typing import Any, Literal, overload

import numpy as np
import numpy.typing as npt
import scipy.interpolate
import xarray as xr

from pycontrails.core import rgi_cython  # type: ignore[attr-defined]

# ------------------------------------------------------------------------------
# Multidimensional interpolation
# ------------------------------------------------------------------------------


logger = logging.getLogger(__name__)


class PycontrailsRegularGridInterpolator(scipy.interpolate.RegularGridInterpolator):
    """Support for performant interpolation over a regular grid.

    This class is a thin wrapper around the
    :class:`scipy.interpolate.RegularGridInterpolator` in order to make typical
    ``pycontrails`` use-cases more efficient.

    #. Avoid ``RegularGridInterpolator`` constructor validation. In :func:`interp`,
    parameters are carefully crafted to fit into the intended form, thereby making
    validation unnecessary.
    #. Override the :meth:`_evaluate_linear` method with a faster implementation. See
    the :meth:`_evaluate_linear` docstring for more information.

    This class should not be used directly. Instead, use the :func:`interp` function.

    .. versionchanged:: 0.40.0

        The :meth:`_evaluate_linear` method now uses a Cython implementation. The dtype
        of the output is now consistent with the dtype of the underlying :attr:`values`

    Parameters
    ----------
    points : tuple[npt.NDArray[np.float64], ...]
        Coordinates of the grid points.
    values : npt.NDArray[np.float_]
        Grid values. The shape of this array must be compatible with the
        coordinates. An error is raised if the dtype is not ``np.float32``
        or ``np.float64``.
    method : str
        Passed into :class:`scipy.interpolate.RegularGridInterpolator`
    bounds_error : bool
        Passed into :class:`scipy.interpolate.RegularGridInterpolator`
    fill_value : float | np.float64 | None
        Passed into :class:`scipy.interpolate.RegularGridInterpolator`
    """

    def __init__(
        self,
        points: tuple[npt.NDArray[np.float64], ...],
        values: npt.NDArray[np.float_],
        method: str,
        bounds_error: bool,
        fill_value: float | np.float64 | None,
    ):
        if values.dtype not in (np.float32, np.float64):
            raise ValueError("values must be a float array")

        self.grid = points
        self.values = values
        self.method = method
        self.bounds_error = bounds_error
        self.fill_value = fill_value

    def _prepare_xi_simple(self, xi: npt.NDArray[np.float64]) -> npt.NDArray[np.bool_] | None:
        """Run looser version of :meth:`_prepare_xi`.

        Parameters
        ----------
        xi : npt.NDArray[np.float64]
            Points at which to interpolate.

        Returns
        -------
        npt.NDArray[np.bool_] | None
            A 1-dimensional Boolean array indicating which points are out of bounds.
            If ``bounds_error`` is ``True``, this will be ``None``, indicating that
            no points are out of bounds. (This is the same convention as
            :meth:`scipy.interpolate.RegularGridInterpolator._prepare_xi`). If
            every point is in bounds, this is set to ``None``.
        """

        if self.bounds_error:
            for i, p in enumerate(xi.T):
                g0 = self.grid[i][0]
                g1 = self.grid[i][-1]
                if not (np.all(p >= g0) and np.all(p <= g1)):
                    raise ValueError(f"One of the requested xi is out of bounds in dimension {i}")

            return None

        out_of_bounds = self._find_out_of_bounds(xi.T)
        if not np.any(out_of_bounds):
            return None
        return out_of_bounds

    def __call__(
        self, xi: npt.NDArray[np.float64], method: str | None = None
    ) -> npt.NDArray[np.float_]:
        """Evaluate the interpolator at the given points.

        Parameters
        ----------
        xi : npt.NDArray[np.float64]
            Points at which to interpolate. Must have shape ``(n, ndim)``, where
            ``ndim`` is the number of dimensions of the interpolator.
        method : str | None
            Override the :attr:`method` to keep parity with the base class.

        Returns
        -------
        npt.NDArray[np.float_]
            Interpolated values. Has shape ``(n,)``. When computing linear interpolation,
            the dtype is the same as the :attr:`values` array.
        """

        method = method or self.method
        if method != "linear":
            return super().__call__(xi, method)

        out_of_bounds = self._prepare_xi_simple(xi)
        xi_indices, norm_distances = self._find_indices(xi.T)

        out = self._evaluate_linear(xi_indices, norm_distances)
        return self._set_out_of_bounds(out, out_of_bounds)

    def _set_out_of_bounds(
        self, out: npt.NDArray[np.float_], out_of_bounds: npt.NDArray[np.bool_] | None
    ) -> npt.NDArray[np.float_]:
        """Set out-of-bounds values to the fill value.

        Parameters
        ----------
        out : npt.NDArray[np.float_]
            Values from interpolation. This is modified in-place.
        out_of_bounds : npt.NDArray[np.bool_] | None
            A 1-dimensional Boolean array indicating which points are out of bounds.

        Returns
        -------
        out : npt.NDArray[np.float_]
            A reference to the ``out`` array.
        """
        if out_of_bounds is not None and self.fill_value is not None:
            out[out_of_bounds] = self.fill_value

        return out

    def _evaluate_linear(
        self,
        indices: npt.NDArray[np.int64],
        norm_distances: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float_]:
        """Evaluate the interpolator using linear interpolation.

        This is a faster alternative to
        :meth:`scipy.interpolate.RegularGridInterpolator._evaluate_linear`.

        .. versionadded:: 0.24

        .. versionchanged:: 0.40.0

            Use Cython routines for evaluating the interpolation when the
            dimension is 1, 2, 3, or 4.

        Parameters
        ----------
        indices : npt.NDArray[np.int64]
            Indices of the grid points to the left of the interpolation points.
            Has shape ``(ndim, n_points)``.
        norm_distances : npt.NDArray[np.float64]
            Normalized distances between the interpolation points and the grid
            points to the left. Has shape ``(ndim, n_points)``.

        Returns
        -------
        npt.NDArray[np.float_]
            Interpolated values with shape ``(n_points,)`` and the same dtype as
            the :attr:`values` array.
        """
        # Let scipy "slow" implementation deal with high-dimensional grids
        if indices.shape[0] > 4:
            return super()._evaluate_linear(indices, norm_distances)

        # Squeeze as much as possible
        # Our cython implementation requires non-degenerate arrays
        non_degen = tuple(s > 1 for s in self.values.shape)
        values = self.values.squeeze()
        indices = indices[non_degen, :]
        norm_distances = norm_distances[non_degen, :]

        ndim, n_points = indices.shape
        out = np.empty(n_points, dtype=self.values.dtype)

        if ndim == 4:
            return rgi_cython.evaluate_linear_4d(values, indices, norm_distances, out)

        if ndim == 3:
            return rgi_cython.evaluate_linear_3d(values, indices, norm_distances, out)

        if ndim == 2:
            return rgi_cython.evaluate_linear_2d(values, indices, norm_distances, out)

        if ndim == 1:
            # np.interp could be better ... although that may also promote the dtype
            return rgi_cython.evaluate_linear_1d(values, indices, norm_distances, out)

        raise ValueError(f"Invalid number of dimensions: {ndim}")


def _floatize_time(
    time: npt.NDArray[np.datetime64], offset: np.datetime64
) -> npt.NDArray[np.float64]:
    """Convert an array of ``np.datetime64`` to an array of ``np.float64``.

    In calls to :class:`scipy.interpolate.RegularGridInterpolator`, it's critical
    that every coordinate be of same type. This creates complications: spatial
    coordinates are float-like, whereas time coordinates are datetime-like. In
    particular, it is not possible to cast an ``np.datetime64`` to a float
    without losing information. In practice, this is not problematic because
    ``np.float64`` has plenty of precision. Previously, this was more of an issue
    because we used ``np.float32``.

    This function uses a fixed time resolution (1 millisecond) to convert the time
    coordinate to a float-like coordinate. The time resolution is taken to avoid
    losing too much information for the time scales we encounter.

    Care is taken to ensure "nat" values are converted to "nan".

    Note that ``xarray`` also must confront this issue. They take a similar approach
    in :func:`xarray.core.missing._floatize_x`. See
    https://github.com/pydata/xarray/blob/d4db16699f30ad1dc3e6861601247abf4ac96567/xarray/core/missing.py#L572

    .. versionchanged:: 0.40.0

        No longer allow the option of converting to ``np.float32``. No longer
        floor the time values to the preceding millisecond.

    Parameters
    ----------
    time : npt.NDArray[np.datetime64]
        Array of ``np.datetime64`` values.
    offset : np.datetime64
        The offset to subtract from ``time``.

    Returns
    -------
    npt.NDArray[np.float64]
        The number of milliseconds since ``offset``.
    """
    delta = time - offset
    resolution = np.timedelta64(1, "ms")
    return delta / resolution


def _localize(da: xr.DataArray, coords: dict[str, np.ndarray]) -> xr.DataArray:
    """Clip ``da`` to the smallest bounding box that contains all of ``coords``.

    Roughly follows approach taken by :func:`xarray.core.missing._localize`. See
    https://github.com/pydata/xarray/blob/56f05c37924071eb4712479d47432aafd4dce38b/xarray/core/missing.py#L557

    Parameters
    ----------
    da : xr.DataArray
        DataArray to clip.
    coords : dict[str, np.ndarray]
        Coordinates to clip to.

    Returns
    -------
    xr.DataArray
        Clipped :class:`xarray.DataArray`. Has the same dimensions as the input ``da``.
        In particular, each dimension of the returned DataArray is a slice of the
        corresponding dimension of the input ``da``.
    """
    indexes: dict[str, Any] = {}
    for dim, arr in coords.items():
        dim_vals = da[dim].values

        # Skip single level
        if dim == "level" and dim_vals.size == 1 and dim_vals.item() == -1:
            continue

        # Create slice
        minval = np.nanmin(arr)
        maxval = np.nanmax(arr)
        imin = np.searchsorted(dim_vals, minval, side="right") - 1
        imin = max(0, imin)
        imax = np.searchsorted(dim_vals, maxval, side="left") + 1
        indexes[dim] = slice(imin, imax)

        # Logging
        n_in_bounds = np.sum((arr >= minval) & (arr <= maxval))
        logger.debug(
            "Interpolation in bounds along dimension %s: %d/%d",
            dim,
            n_in_bounds,
            arr.size,
        )

    return da.isel(**indexes)


@overload
def interp(
    longitude: npt.NDArray[np.float_],
    latitude: npt.NDArray[np.float_],
    level: npt.NDArray[np.float_],
    time: npt.NDArray[np.datetime64],
    da: xr.DataArray,
    method: str,
    bounds_error: bool,
    fill_value: float | np.float64 | None,
    localize: bool,
    *,
    indices: RGIArtifacts | None = ...,
    return_indices: Literal[False] = ...,
) -> npt.NDArray[np.float_]:
    ...


@overload
def interp(
    longitude: npt.NDArray[np.float_],
    latitude: npt.NDArray[np.float_],
    level: npt.NDArray[np.float_],
    time: npt.NDArray[np.datetime64],
    da: xr.DataArray,
    method: str,
    bounds_error: bool,
    fill_value: float | np.float64 | None,
    localize: bool,
    *,
    indices: RGIArtifacts | None = ...,
    return_indices: Literal[True],
) -> tuple[npt.NDArray[np.float_], RGIArtifacts]:
    ...


@overload
def interp(
    longitude: npt.NDArray[np.float_],
    latitude: npt.NDArray[np.float_],
    level: npt.NDArray[np.float_],
    time: npt.NDArray[np.datetime64],
    da: xr.DataArray,
    method: str,
    bounds_error: bool,
    fill_value: float | np.float64 | None,
    localize: bool,
    *,
    indices: RGIArtifacts | None = ...,
    return_indices: bool = ...,
) -> npt.NDArray[np.float_] | tuple[npt.NDArray[np.float_], RGIArtifacts]:
    ...


def interp(
    longitude: npt.NDArray[np.float_],
    latitude: npt.NDArray[np.float_],
    level: npt.NDArray[np.float_],
    time: npt.NDArray[np.datetime64],
    da: xr.DataArray,
    method: str,
    bounds_error: bool,
    fill_value: float | np.float64 | None,
    localize: bool,
    *,
    indices: RGIArtifacts | None = None,
    return_indices: bool = False,
) -> npt.NDArray[np.float_] | tuple[npt.NDArray[np.float_], RGIArtifacts]:
    """Interpolate over a grid with ``localize`` option.

    .. versionchanged:: 0.25.6

        Utilize scipy 1.9 upgrades to remove singleton dimensions.

    .. versionchanged:: 0.26.0

        Include ``indices`` and ``return_indices`` experimental parameters.
        Currently, nan values in ``longitude``, ``latitude``, ``level``, or ``time``
        are always propagated through to the output, regardless of ``bounds_error``.
        In other words, a ValueError for an out of bounds coordinate is only raised
        if a non-nan value is out of bounds.

    .. versionchanged:: 0.40.0

        When ``return_indices`` is True, an instance of :class:`RGIArtifacts`
        is used to store the indices artifacts.

    Parameters
    ----------
    longitude, latitude, level, time : np.ndarray
        Coordinates of points to be interpolated. These parameters have the same
        meaning as ``x`` in analogy with :func:`numpy.interp`. All four of these
        arrays must be 1 dimensional of the same size.
    da : xr.DataArray
        Gridded data interpolated over. Must adhere to ``MetDataArray`` conventions.
        In particular, the dimensions of ``da`` must be ``longitude``, ``latitude``,
        ``level``, and ``time``. The three spatial dimensions must be monotonically
        increasing with ``float64`` dtype. The ``time`` dimension must be
        monotonically increasing with ``datetime64`` dtype.
        Assumed to be cheap to load into memory (:attr:`xr.DataArray.values` is
        used without hesitation).
    method : str
        Passed into :class:`scipy.interpolate.RegularGridInterpolator`.
    bounds_error : bool
        Passed into :class:`scipy.interpolate.RegularGridInterpolator`.
    fill_value : float | np.float64 | None
        Passed into :class:`scipy.interpolate.RegularGridInterpolator`.
    localize : bool
        If True, clip ``da`` to the smallest bounding box that contains all of
        ``coords``.
    indices : tuple | None, optional
        Experimental. Provide intermediate artifacts computed by
        :meth:``scipy.interpolate.RegularGridInterpolator._find_indices`
        to avoid redundant computation. If known and provided, this can speed
        up interpolation by avoiding an unnecessary call to ``_find_indices``.
        By default, None. Must be used precisely.
    return_indices : bool, optional
        If True, return output of :meth:`scipy.interpolate.RegularGridInterpolator._find_indices`
        in addition to interpolated values.

    Returns
    -------
    npt.NDArray[np.float_] | tuple[npt.NDArray[np.float_], RGIArtifacts]
        Interpolated values with same size as ``longitude``. If ``return_indices``
        is True, return intermediate indices artifacts as well.

    See Also
    --------
    - :meth:`MetDataArray.interpolate`
    - :func:`scipy.interpolate.interpn`
    - :class:`scipy.interpolate.RegularGridInterpolator`
    """
    if localize:
        coords = {"longitude": longitude, "latitude": latitude, "level": level, "time": time}
        da = _localize(da, coords)

    # Using da.coords.variables is slightly more performant than da["longitude"].values
    x = da.coords.variables["longitude"].values
    y = da.coords.variables["latitude"].values
    z = da.coords.variables["level"].values
    if any(v.dtype != np.float64 for v in (x, y, z)):
        raise ValueError(
            "da must have float64 dtype for longitude, latitude, and level coordinates"
        )

    # Convert t and time to float64
    t = da.coords.variables["time"].values
    offset = t[0]
    t = _floatize_time(t, offset)

    single_level = z.size == 1 and z.item() == -1
    points: tuple[npt.NDArray[np.float_], ...]
    if single_level:
        values = da.values.squeeze(axis=2)
        points = x, y, t
    else:
        values = da.values
        points = x, y, z, t

    interp_ = PycontrailsRegularGridInterpolator(
        points=points,
        values=values,
        method=method,
        bounds_error=bounds_error,
        fill_value=fill_value,
    )

    if indices is None:
        xi = _buildxi(longitude, latitude, level, time, offset, single_level)
        if return_indices:
            out, indices = _linear_interp_with_indices(interp_, xi, localize, None)
            return out, indices
        return interp_(xi)

    out, indices = _linear_interp_with_indices(interp_, None, localize, indices)
    if return_indices:
        return out, indices
    return out


def _buildxi(
    longitude: npt.NDArray[np.float_],
    latitude: npt.NDArray[np.float_],
    level: npt.NDArray[np.float_],
    time: npt.NDArray[np.datetime64],
    offset: np.datetime64,
    single_level: bool,
) -> npt.NDArray[np.float64]:
    """Build the input array for interpolation.

    The implementation below achieves the same result as the following::

        np.stack([longitude, latitude, level, time_float], axis=1])

    This implementation is slightly faster than the above.
    """

    time_float = _floatize_time(time, offset)

    if single_level:
        ndim = 3
    else:
        ndim = 4

    shape = longitude.size, ndim
    xi = np.empty(shape, dtype=np.float64)

    xi[:, 0] = longitude
    xi[:, 1] = latitude
    if not single_level:
        xi[:, 2] = level
    xi[:, -1] = time_float

    return xi


def _linear_interp_with_indices(
    interp: PycontrailsRegularGridInterpolator,
    xi: npt.NDArray[np.float64] | None,
    localize: bool,
    indices: RGIArtifacts | None,
) -> tuple[npt.NDArray[np.float_], RGIArtifacts]:
    if interp.method != "linear":
        raise ValueError("Parameter 'indices' is only supported for 'method=linear'")
    if localize:
        raise ValueError("Parameter 'indices' is only supported for 'localize=False'")

    if indices is None:
        assert xi is not None, "xi must be provided if indices is None"
        out_of_bounds = interp._prepare_xi_simple(xi)
        xi_indices, norm_distances = interp._find_indices(xi.T)
        indices = RGIArtifacts(xi_indices, norm_distances, out_of_bounds)

    out = interp._evaluate_linear(indices.xi_indices, indices.norm_distances)
    out = interp._set_out_of_bounds(out, indices.out_of_bounds)
    return out, indices


@dataclasses.dataclass
class RGIArtifacts:
    """An interface to intermediate RGI interpolation artifacts."""

    xi_indices: npt.NDArray[np.int64]
    norm_distances: npt.NDArray[np.float64]
    out_of_bounds: npt.NDArray[np.bool_] | None


# ------------------------------------------------------------------------------
# 1 dimensional interpolation
# ------------------------------------------------------------------------------


class EmissionsProfileInterpolator:
    """Support for interpolating a profile on a linear or logarithmic scale.

    This class simply wraps :func:`numpy.interp` with fixed values for the
    ``xp`` and ``fp`` arguments. Unlike :class:`xarray.DataArray` interpolation,
    the `numpy.interp` automatically clips values outside the range of the
    ``xp`` array.

    Parameters
    ----------
    xp : npt.NDarray[np.float_]
        Array of x-values. These must be strictly increasing and free from
        any nan values. Passed to :func:`numpy.interp`.
    fp : npt.NDarray[np.float_]
        Array of y-values. Passed to :func:`numpy.interp`.
    drop_duplicates : bool, optional
        Whether to drop duplicate values in ``xp``. Defaults to ``True``.

    Examples
    --------
    >>> xp = np.array([3, 7, 10, 30], dtype=float)
    >>> fp = np.array([0.1, 0.2, 0.3, 0.4], dtype=float)
    >>> epi = EmissionsProfileInterpolator(xp, fp)
    >>> # Interpolate a single value
    >>> epi.interp(5)
    0.150000...

    >>> # Interpolate a single value on a logarithmic scale
    >>> epi.log_interp(5)
    1.105171...

    >>> # Demonstrate speed up compared with xarray.DataArray interpolation
    >>> import time, xarray as xr
    >>> da = xr.DataArray(fp, dims=["x"], coords={"x": xp})

    >>> inputs = [np.random.uniform(0, 31, size=200) for _ in range(1000)]
    >>> t0 = time.perf_counter()
    >>> xr_out = [da.interp(x=x.clip(3, 30)).values for x in inputs]
    >>> t1 = time.perf_counter()
    >>> np_out = [epi.interp(x) for x in inputs]
    >>> t2 = time.perf_counter()
    >>> assert np.allclose(xr_out, np_out)

    >>> # We see a 100 fold speed up (more like 500x faster, but we don't
    >>> # want the test to fail!)
    >>> assert t2 - t1 < (t1 - t0) / 100
    """

    def __init__(
        self, xp: npt.NDArray[np.float_], fp: npt.NDArray[np.float_], drop_duplicates: bool = True
    ) -> None:
        if drop_duplicates:
            # Using np.diff to detect duplicates ... this assumes xp is sorted.
            # If xp is not sorted, an ValueError will be raised in _validate
            mask = np.abs(np.diff(xp, prepend=np.inf)) < 1e-15  # small tolerance
            xp = xp[~mask]
            fp = fp[~mask]

        self.xp = np.asarray(xp)
        self.fp = np.asarray(fp)
        self._validate()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(xp={self.xp}, fp={self.fp})"

    def _validate(self) -> None:
        if not len(self.xp):
            raise ValueError("xp must not be empty")
        if len(self.xp) != len(self.fp):
            raise ValueError("xp and fp must have the same length")
        if not np.all(np.diff(self.xp) > 0.0):
            raise ValueError("xp must be strictly increasing")
        if np.any(np.isnan(self.xp)):
            raise ValueError("xp must not contain nan values")

    def interp(self, x: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
        """Interpolate x against xp and fp.

        Parameters
        ----------
        x : npt.NDArray[np.float_]
            Array of x-values to interpolate.

        Returns
        -------
        npt.NDArray[np.float_]
            Array of interpolated y-values arising from the x-values. The ``dtype`` of
            the output array is the same as the ``dtype`` of ``x``.
        """
        # Need to explicitly cast back to x.dtype
        # https://github.com/numpy/numpy/issues/11214
        x = np.asarray(x)
        dtype = np.result_type(x, np.float32)
        return np.interp(x, self.xp, self.fp).astype(dtype)

    def log_interp(self, x: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
        """Interpolate x against xp and fp on a logarithmic scale.

        This method composes the following three functions.
            1. :func:`numpy.log`
            2. :meth:`interp`
            3. :func:`numpy.exp`

        Parameters
        ----------
        x : npt.NDArray[np.float_]
            Array of x-values to interpolate.

        Returns
        -------
        npt.NDArray[np.float_]
            Array of interpolated y-values arising from the x-values.
        """
        return np.exp(self.interp(np.log(x)))



"""
Benchmark test data for `Cocip` Model.

See README.md for setup instructions.
"""

from __future__ import annotations

import logging
import pathlib

import pandas as pd

from pycontrails import Fleet, Flight, MetDataset
from pycontrails.datalib.ecmwf import ERA5
from pycontrails.ext.bada import BADAFlight
from pycontrails.models.cocip import Cocip
from pycontrails.models.humidity_scaling import ExponentialBoostLatitudeCorrectionHumidityScaling
from pycontrails.physics import units

# set up logging
LOG = logging.getLogger("pycontrails")
logging.basicConfig(
    # filename="benchmark.log",
    format="%(asctime)s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)

# setup
INPUT_PATH = pathlib.Path("inputs")
OUTPUT_PATH = pathlib.Path("outputs")
FLIGHT_METADATA_FILE = "flight-metadata.pq"
FLIGHT_WAYPOINT_FILE = "flight-waypoints.pq"

# ----
# Load Flights
# ----


def load_flights() -> list[Flight]:
    """Load flight data from input file.

    Returns
    -------
    list[Flight]
        List of Flight objects loaded from csv file
    """
    LOG.info("Opening flight data")

    # load flight metadata
    df_flights_metadata = pd.read_parquet(INPUT_PATH / "flight" / FLIGHT_METADATA_FILE)
    df_flights_metadata.set_index("flight_id", inplace=True)

    # Load flight waypoints
    df_flights_waypoints = pd.read_parquet(INPUT_PATH / "flight" / FLIGHT_WAYPOINT_FILE)

    # get altitude in m
    df_flights_waypoints["altitude"] = units.ft_to_m(df_flights_waypoints["altitude_ft"])

    # load datetimes
    df_flights_waypoints["time"] = pd.to_datetime(df_flights_waypoints["time"])

    cols_wypt = ["flight_id", "longitude", "latitude", "altitude", "time"]
    flights = []

    # load metadata and create Flight class
    for flt_id, df_flt_wypts in df_flights_waypoints.groupby("flight_id"):
        # Aircraft and engine properties
        attrs = {
            "flight_id": flt_id,
            "aircraft_type": df_flights_metadata.loc[flt_id]["aircraft_type_icao"],
            "engine_name": df_flights_metadata.loc[flt_id]["engine_type_edb"],
            "engine_uid": df_flights_metadata.loc[flt_id]["engine_uid"],
            "load_factor": df_flights_metadata.loc[flt_id]["assumed_load_factor"],
        }

        # Create flight object
        fl = Flight(data=df_flt_wypts[cols_wypt], attrs=attrs)
        flights.append(fl)

    return flights


# ----
# Load ERA5
# ----


def load_ERA5() -> tuple[MetDataset, MetDataset]:
    """Load meteorology and radiation datasets.

    Returns
    -------
    tuple(MetDataset, MetDataset):
        Meteorology on pressure levels, Radiation on single level
    """
    LOG.info("Opening meteorology and radiation datasets")

    time = ("2020-01-01 00:00:00", "2020-01-02 08:00:00")
    pressure_levels = [
        1000,
        975,
        950,
        925,
        900,
        875,
        850,
        825,
        800,
        775,
        750,
        700,
        650,
        600,
        550,
        500,
        450,
        400,
        350,
        300,
        250,
        225,
        200,
        175,
        150,
        125,
        100,
    ]

    # load local ERA5 files for this domain
    # Note the files at "paths=" must be directly from ECMWF (not processed by pycontrails first)
    era5pl = ERA5(
        time=time,
        variables=Cocip.met_variables,
        pressure_levels=pressure_levels,
        paths=list(pathlib.Path(INPUT_PATH / "met").glob("*.nc")),
        cachestore=None,
    )
    era5sl = ERA5(
        time=time,
        variables=Cocip.rad_variables,
        paths=list(pathlib.Path(INPUT_PATH / "rad").glob("*.nc")),
        cachestore=None,
    )

    # create `MetDataset` from sources
    met = era5pl.open_metdataset(xr_kwargs=dict(parallel=False, format="NETCDF3_CLASSIC"))
    rad = era5sl.open_metdataset(xr_kwargs=dict(parallel=False, format="NETCDF3_CLASSIC"))

    return met, rad


# ----
# Cocip
# ----


def run_cocip() -> Cocip:
    """Run Cocip model on flights.

    Returns
    -------
    Cocip
        Evaluated cocip model
    """

    flights = load_flights()
    met, rad = load_ERA5()

    # create Aircraft Performance model
    aircraft_performance = BADAFlight(
        bada3_path=INPUT_PATH / ".." / "bada" / "bada3",
        bada4_path=INPUT_PATH / ".." / "bada" / "bada4",
        copy_source=False,
        n_iter=3,
    )

    # create Humidity Scaling model
    humidity_scaling = ExponentialBoostLatitudeCorrectionHumidityScaling(copy_source=False)

    params = {
        "max_age": pd.Timedelta(12, "h"),
        "dt_integration": pd.Timedelta(5, "m"),
        "radiative_heating_effects": True,
        "max_seg_length_m": 60_000,
        "aircraft_performance": aircraft_performance,
        "humidity_scaling": humidity_scaling,
        "process_emissions": True,
        "verbose_outputs": True,
    }

    cocip = Cocip(
        met=met,
        rad=rad,
        params=params,
    )

    LOG.info("Evaluating Cocip model")

    # returns list of Flight outputs
    # output_flights = cocip.eval(source=flights[14:15])
    _ = cocip.eval(source=flights)

    return cocip


# -------
# Results
# -------


def parse_flight_results(fleet: Fleet) -> pd.DataFrame:
    """Interpret fleet results from Cocip evaluation.

    Uses `Cocip.flight` output instead of list[Flight]

    Parameters
    ----------
    fleet : Fleet
        Cocip.flight output after Cocip.eval(...)

    Returns
    -------
    pd.DataFrame
    """
    # add extra dimensions
    fleet["altitude_ft"] = fleet.altitude_ft
    fleet["air_pressure"] = fleet.air_pressure
    # fleet["mach_number"] = units.tas_to_mach_number(
    #     fleet["true_airspeed"], fleet["air_temperature"]
    # )

    # create dataframe from fleet
    fleet_df = fleet.dataframe

    # convert timestamps into ints (unix epoch)
    fleet_df["time_unix"] = (
        pd.to_datetime(fleet["time"]) - pd.Timestamp("1970-01-01")
    ) // pd.Timedelta("1s")

    # convert timedelta into seconds
    fleet_df["contrail_age"] = fleet["contrail_age"] // pd.Timedelta("1s")

    # drop some excess columns
    fleet_df = fleet_df.drop(
        columns=[
            "co_ei",
            "hc_ei",
            "nox_ei",
            "nvpm_mass",
            "nvpm_ei_m",
            "nvpm_number",
            "thrust_setting",
        ]
    )

    # fill na with -9999
    # fleet_df = fleet_df.fillna(value=-9999)

    return fleet_df


def parse_contrail_results(contrail: pd.DataFrame) -> pd.DataFrame:
    """Interpret contrail results from Cocip evaluation.

    Uses `Cocip.contrail` output

    Parameters
    ----------
    contrail : pd.DataFrame
        Cocip.contrail output after Cocip.eval(...)

    Returns
    -------
    pd.DataFrame
    """

    contrail = contrail.reset_index(drop=True)

    # convert timestamps into ints (unix epoch)
    contrail["time_unix"] = (
        pd.to_datetime(contrail["time"]) - pd.Timestamp("1970-01-01")
    ) // pd.Timedelta("1s")
    contrail["formation_time_unix"] = (
        pd.to_datetime(contrail["formation_time"]) - pd.Timestamp("1970-01-01")
    ) // pd.Timedelta("1s")

    # convert timedelta into seconds
    contrail["dt_integration"] = contrail["dt_integration"] // pd.Timedelta("1s")
    contrail["age"] = contrail["age"] // pd.Timedelta("1s")

    # cast bools to ints
    contrail.loc[:, ["continuous", "persistent"]] = contrail.loc[
        :, ["continuous", "persistent"]
    ].astype(int)

    contrail = contrail.drop(columns=["age_hours", "cumul_heat", "cumul_differential_heat"])

    # fill na with -9999
    # contrail = contrail.fillna(value=-9999)

    return contrail



"""Unit tests for the pycontrails.core.polygons.py module."""

from __future__ import annotations

import numpy as np
import pytest
import shapely

from pycontrails import MetDataArray
from pycontrails.core import polygon


@pytest.fixture
def carr() -> np.ndarray:
    """Fixture with continuous padded data."""
    rng = np.random.default_rng(654)
    arr = rng.uniform(0, 0.8, size=(100, 100))
    arr[25:35, 55:70] = 1  # big island
    arr[5:10, 5:10] = 1  # small island
    return np.pad(arr, 1, constant_values=0)


@pytest.fixture
def barr() -> np.ndarray:
    """Fixture with binary padded data."""
    rng = np.random.default_rng(654)
    arr = rng.integers(0, 2, size=(100, 100)).astype(float)
    return np.pad(arr, 1, constant_values=0)


@pytest.fixture
def hawaiian_earrings() -> np.ndarray:
    """Fixture reminiscent of Hawaiian earrings."""
    img = np.zeros((10, 10))

    sl = slice(1, 9)
    img[sl, sl] = 1

    sl = slice(2, 8)
    img[sl, sl] = 0

    sl = slice(3, 7)
    img[sl, sl] = 1

    sl = slice(4, 6)
    img[sl, sl] = 0

    return img


@pytest.mark.parametrize("arr_name", ["carr", "barr", "hawaiian_earrings"])
def test_find_multipolygon_depth1(arr_name: str, request: pytest.FixtureRequest):
    """Test the `polygon.find_multipolygon` function at depth 1."""

    arr = request.getfixturevalue(arr_name)
    assert isinstance(arr, np.ndarray)

    mp = polygon.find_multipolygon(
        arr,
        threshold=0.5,
        min_area=0,
        epsilon=0.1,
        interiors=False,
    )
    assert isinstance(mp, shapely.MultiPolygon)
    assert mp.is_valid

    # All polygons have correct orientation and no holes
    n_polys = 0
    for p in mp.geoms:
        assert isinstance(p, shapely.Polygon)
        assert p.is_valid
        assert not p.has_z
        assert p.is_simple
        assert p.exterior.is_ccw
        assert not p.interiors
        assert p.area
        assert p.length
        n_polys += 1

    # pin the number of children
    if arr_name == "carr":
        assert n_polys == 227
    elif arr_name == "barr":
        assert n_polys == 23
    elif arr_name == "hawaiian_earrings":
        assert n_polys == 1


@pytest.mark.parametrize(
    "min_area,n_children ",
    [(1, 144), (2, 100), (3, 87), (4, 76), (6, 58), (12, 36), (20, 27)],
)
def test_find_multipolygon_min_area(carr: np.ndarray, min_area: float, n_children: int):
    """Show that increasing min_area decreases the number of children."""
    mp = polygon.find_multipolygon(
        carr,
        threshold=0.5,
        min_area=min_area,
        epsilon=0.1,
        interiors=False,
    )
    n_polys = len(mp.geoms)
    assert n_polys == n_children
    for p in mp.geoms:
        assert p.area >= min_area
        assert not p.interiors


@pytest.mark.parametrize("arr_name", ["barr", "hawaiian_earrings"])
def test_find_multipolygons_with_interiors(arr_name: str, request: pytest.FixtureRequest):
    """Test find_multipolygon with interiors=True."""

    arr = request.getfixturevalue(arr_name)
    assert isinstance(arr, np.ndarray)

    mp = polygon.find_multipolygon(
        arr,
        threshold=0.02,
        min_area=0,
        epsilon=0.1,
        interiors=True,
    )

    interior_found = False
    assert isinstance(mp, shapely.MultiPolygon)
    assert mp.is_valid
    for p in mp.geoms:
        assert p.is_valid
        interior_found = interior_found or p.interiors
    assert interior_found


@pytest.mark.filterwarnings("ignore:Longitude and latitude are not evenly spaced")
@pytest.mark.parametrize("altitude", [None, 10001.2])
def test_multipolygon_to_geojson(barr: np.ndarray, altitude: float | None):
    """Test the `multipolygon_to_geojson` function."""
    sl = slice(1, -1)
    mp = polygon.find_multipolygon(
        barr[sl, sl],
        threshold=0.5,
        min_area=0,
        epsilon=0.1,
        convex_hull=False,
        longitude=np.arange(0, 25, 0.25),
        latitude=np.arange(-75, -25, 0.5),
    )
    assert len(mp.geoms) == 23

    geojson = polygon.multipolygon_to_geojson(mp, altitude=altitude)
    assert isinstance(geojson, dict)
    assert geojson["type"] == "Feature"
    assert geojson["geometry"]["type"] == "MultiPolygon"
    assert len(geojson["geometry"]["coordinates"]) == 23
    for poly in geojson["geometry"]["coordinates"]:
        for ring in poly:
            contour = np.array(ring)
            assert contour.ndim == 2

            assert np.all(contour[:, 0] <= 25.25)
            assert np.all(contour[:, 0] >= -0.5)
            assert np.all(contour[:, 1] <= -25.0)
            assert np.all(contour[:, :1] >= -75.5)

            if altitude is not None:
                assert contour.shape[1] == 3
                assert np.all(contour[:, 2] == altitude)
            else:
                assert contour.shape[1] == 2


@pytest.mark.parametrize("interiors", [True, False])
def test_contour_find_islands(carr: np.ndarray, interiors: bool):
    """Confirm both large islands are found.

    See fixture definition.
    """
    mp = polygon.find_multipolygon(
        carr,
        threshold=0.9,
        min_area=0,
        epsilon=0.05,
        interiors=interiors,
    )
    n_polys = len(mp.geoms)
    assert n_polys == 2
    poly1, poly2 = mp.geoms
    assert not poly1.interiors
    assert not poly2.interiors
    assert len(poly1.exterior.coords) == 9
    assert len(poly2.exterior.coords) == 9

    assert poly1.exterior.is_ccw
    assert poly2.exterior.is_ccw
    assert poly1.area == 149.5
    assert poly2.area == 24.5

    # with high enough min_area, only one island is found
    mp = polygon.find_multipolygon(
        carr,
        threshold=0.9,
        min_area=25,
        epsilon=0.1,
        interiors=interiors,
    )
    assert len(mp.geoms) == 1

    # and if min_area is even higher, no islands are found
    mp = polygon.find_multipolygon(
        carr,
        threshold=0.9,
        min_area=150,
        epsilon=0.1,
        interiors=interiors,
    )
    assert len(mp.geoms) == 0


def test_hawaiian_earrings(hawaiian_earrings: np.ndarray):
    """Show that `find_multipolygon` can find deeply nested contours."""

    mp = polygon.find_multipolygon(hawaiian_earrings, threshold=0.5, min_area=0, epsilon=0.1)
    assert isinstance(mp, shapely.MultiPolygon)
    assert mp.is_valid
    assert len(mp.geoms) == 2

    for poly in mp.geoms:
        assert isinstance(poly, shapely.Polygon)
        assert poly.is_valid
        exterior = poly.exterior
        assert exterior.is_ccw
        assert len(exterior.coords) == 9

        assert len(poly.interiors) == 1
        interior = poly.interiors[0]
        assert isinstance(interior, shapely.LinearRing)
        assert not interior.is_ccw
        assert len(interior.coords) == 9


@pytest.mark.parametrize("iso_value", [3e8, 6e8, 1e9])
@pytest.mark.parametrize("min_area", [0.0, 1.0, 2.0])
def test_polygon_roundoff_error(polygon_bug: MetDataArray, iso_value: float, min_area: float):
    """Confirm that a polygon bug is fixed.

    On pycontrails 0.32.0, rounding polygon vertices caused a bug where the
    second and second to last vertices were the same. This test confirms that
    the bug is fixed.
    """
    assert polygon_bug.shape == (720, 261, 1, 1)

    feature = polygon_bug.to_polygon_feature(iso_value=iso_value, min_area=min_area)
    coords = feature["geometry"]["coordinates"]
    for polys in coords:
        for i, ring in enumerate(polys):
            lr = shapely.LinearRing(ring)
            assert lr.is_valid
            assert lr.is_simple
            assert lr.is_ring
            assert lr.is_closed
            assert lr.is_ccw == (i == 0)

            assert np.array_equal(ring[0], ring[-1])
            assert not np.array_equal(ring[1], ring[-2])
            # Check that no other vertices are duplicated
            assert len(np.unique(ring, axis=0)) == len(ring) - 1



"""
Meteorology and Air Traffic data source wrappers.

See individual modules for met variables and additional exports.

- :module:`pycontrails.datalib.ecmwf`
- :module:`pycontrails.datalib.gfs`
- :module:`pycontrails.datalib.spire`
"""



"""
Compare benchmark test outputs with previous outputs.

```
$ python compare.py
```
"""

import logging

import benchmark
import numpy as np
import pandas as pd

# relative tolerances for comparison
FLIGHT_TOLERANCE = 1e-6
CONTRAIL_TOLERANCE = 1e-3

# logging set up in benchmark.py
LOG = logging.getLogger("pycontrails")

LOG.info("Running Cocip over flights")

# run model on benchmark inputs
cocip = benchmark.run_cocip()

LOG.info("Comparing outputs with previous results")


# -----------------
# Flight comparison
# -----------------

fleet = benchmark.parse_flight_results(cocip.source)

for fid, df in fleet.groupby("flight_id"):
    df = df.reset_index(drop=True)

    # read previous results
    df_prev = pd.read_parquet(benchmark.OUTPUT_PATH / "flight" / f"{fid}.pq")

    # test all number types
    for col in df.select_dtypes(include=np.number):
        np.testing.assert_allclose(
            df_prev[col].to_numpy(), df[col].to_numpy(), rtol=FLIGHT_TOLERANCE
        )

LOG.info(f"Successfully compared {len(fleet)} flight rows")

# -------------------
# Contrail comparison
# -------------------


contrail = benchmark.parse_contrail_results(cocip.contrail)

for fid, df in contrail.groupby("flight_id"):
    df = df.reset_index(drop=True)

    # read previous results
    df_prev = pd.read_parquet(benchmark.OUTPUT_PATH / "contrail" / f"{fid}.pq")

    # test all number types
    for col in df.select_dtypes(include=np.number):
        np.testing.assert_allclose(
            df_prev[col].to_numpy(), df[col].to_numpy(), rtol=CONTRAIL_TOLERANCE
        )

LOG.info(f"Successfully compared {len(contrail)} contrail rows")



"""Support for building pycontrails Cython extension modules.

See https://stackoverflow.com/q/66157987/ for details.
"""

import numpy
import setuptools
from Cython.Build import cythonize

rgi_ext = setuptools.Extension(
    "pycontrails.core.rgi_cython",
    ["pycontrails/core/rgi_cython.pyx"],
    language="c",
    include_dirs=[numpy.get_include()],
)


setuptools.setup(ext_modules=cythonize([rgi_ext]))



"""
Monkey patch deprecated functions for validation script
"""

from __future__ import annotations

import numpy as np
import xarray as xr

from pycontrails import JetA, MetDataset
from pycontrails.models import sac
from pycontrails.models.tau_cirrus import _assign_attrs, cirrus_effective_extinction_coef
from pycontrails.physics import constants, geo, thermo, units
from pycontrails.utils.types import ArrayLike

# ----------
# geo module
# ----------


def segment_angle(longitude: np.ndarray, latitude: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Calculates the sin(a) and cos(a) for the
    angle between each segment and the longitudinal axis


                (lon_2, lat_2)  X
                               /|
                              / |
                             /  |
                            /   |
                           /    |
                          /     |
                         /      |
        (lon_1, lat_1)  X -------> longitude (x-axis)

    Returns
    -------
    np.ndarray, np.ndarray
        sin(a), cos(a), where a is the angle between the segment and the longitudinal axis

    See Also
    --------
    :func:`haversine`
    """
    lats_next = np.roll(latitude, -1)
    lats_avg = 0.5 * (latitude + lats_next)

    # append nan to ensure length is the same as the original waypoints
    d_lon = np.append(np.diff(longitude), np.nan)
    d_lat = np.append(np.diff(latitude), np.nan)

    dist = geo.segment_haversine(longitude, latitude)
    sin_a = (constants.radius_earth * units.degrees_to_radians(d_lat)) / dist
    sin_a[-1] = 0  # last element is a discontinuous segment
    assert isinstance(sin_a, np.ndarray)

    cos_a = (
        constants.radius_earth
        * units.degrees_to_radians(d_lon)
        * np.cos(units.degrees_to_radians(lats_avg))
        / dist
    )
    cos_a[-1] = 1  # last element is a discontinuous segment

    return sin_a, cos_a


# --------------------------
# contrail_properties module
# --------------------------


def segment_length_ratio(
    seg_length: np.ndarray, seg_length_t: np.ndarray, continuous: np.ndarray
) -> np.ndarray:
    """
    Calculates the ratio of contrail segment length before
    and after it is advected to the new location.

    TODO: @roger - I don't quite understand this function
    TODO: currently only works for 1D np.ndarray

    Parameters
    ----------
    seg_length : np.ndarray
        Original segment length of contrail waypoint, [:math:`m`]
    seg_length_t : np.ndarray
        Segment length of contrail waypoint after time step and advection, [:math:`m`]
    continuous : np.ndarray
        Description

    Returns
    -------
    np.ndarray
        Segment length ratio before and after it is advected to the new location.
    """

    # init new output
    seg_ratio = np.ones(seg_length_t.shape)

    s = (seg_length / np.maximum(seg_length_t, 1)) ** 0.5
    s_next = np.roll(s, 1)
    s_next[0] = 0

    # TODO: @roger what is going on here? Seems like something is duplicated
    bool_ops = continuous & (s > 0)
    seg_ratio[bool_ops] = seg_ratio[bool_ops] * s[bool_ops]
    bool_ops = s_next > 0
    seg_ratio[bool_ops] = seg_ratio[bool_ops] * s_next[bool_ops]

    return seg_ratio


def mean_energy_flux_per_m(
    rad_flux_per_m: np.ndarray,
    dt: np.ndarray | np.timedelta64,
    *,
    continuous: np.ndarray | None = None,
) -> np.ndarray:
    """
    Calculates the mean energy flux per length of contrail between two waypoints.

    Parameters
    ----------
    rad_flux_per_m : np.ndarray
        Mean radiative flux between time steps for waypoint, [:math:`W m^{-1}`].
        See :func:`mean_radiative_flux_per_m`.
    dt : np.ndarray
        timedelta of timestep for each waypoint

    Returns
    -------
    np.ndarray
        Mean energy flux per length of contrail between two waypoints, [:math:`J m^{-1}`]
    """
    # convert dt to seconds value
    dt_s = dt / np.timedelta64(1, "s")

    # TODO: Average between waypoints? Output energy_flux_0 in the future?
    energy_flux = rad_flux_per_m * dt_s
    energy_flux_t = np.roll(energy_flux, -1)
    energy_flux_per_m = (energy_flux + energy_flux_t) * 0.5
    if continuous is not None:
        energy_flux_per_m[~continuous] = 0

    return energy_flux_per_m


# ----------
# tau cirrus
# ----------


def tau_cirrus_original(met: MetDataset) -> xr.DataArray:
    """Calculate the optical depth of NWP cirrus around each pressure level.

    This implementation of `tau_cirrus` more closely aligned with Schumann's Fortran implementation.
    See `tau_cirrus` and check the `test_tau_cirrus` unit test module for agreement.
    """

    da = xr.zeros_like(met.data["specific_cloud_ice_water_content"])

    levels = met.data["level"]
    geopotential_height = met.data["geopotential"] / constants.g  # m

    for z in np.arange(len(levels)):
        z_top = max(z - 1, 0)  # Top Layer Index
        z_btm = min(z + 1, len(levels) - 1)  # Bottom Layer Index
        ds = met.data[dict(level=z)]

        beta_e = cirrus_effective_extinction_coef(
            ds["specific_cloud_ice_water_content"],
            ds["air_temperature"],
            ds["air_pressure"],
        )

        dz = 0.5 * (geopotential_height[dict(level=z_top)] - geopotential_height[dict(level=z_btm)])

        da[dict(level=z)] = da[dict(level=z_top)] + (beta_e * dz)

    return _assign_attrs(da)


# ------
# Units
# ------


def pl_to_m(pl: np.ndarray) -> np.ndarray:
    r"""Convert from pressure level (hPa) to altitude (m).

    Parameters
    ----------
    p : ArrayLike
        pressure level, [:math:`hPa`], [:math:`mbar`]

    Returns
    -------
    ArrayLike
        altitude, [:math:`m`]

    Notes
    -----
    Source: https://ozreport.com/docs/MarkGrahamarticle.pdf
    """

    pl_pa = pl * 100
    return (constants.T_msl / 0.0065) * (1 - (pl_pa / constants.p_surface) ** (1 / 5.255))


def _m_to_T_isa(h: np.ndarray) -> np.ndarray:
    """Calculate the ambient temperature (K) for a given altitude (m).

    Assumes the ICAO standard atmosphere.


    Parameters
    ----------
    h : ArrayLike
        altitude, [:math:`m`]

    Returns
    -------
    ArrayLike
        ICAO standard atmosphere ambient temperature, [:math:`K`]

    Notes
    -----
    Source: https://en.wikipedia.org/wiki/International_Standard_Atmosphere#ICAO_Standard_Atmosphere
    """
    T_isa = constants.T_msl + (h * constants.T_lapse_rate)
    T_stratosphere_isa = constants.T_msl + (constants.T_lapse_rate * constants.h_tropopause)

    # array
    if isinstance(T_isa, (np.ndarray, xr.DataArray)):
        # If h is ArrayLike, it supports indexing
        T_isa[h > constants.h_tropopause] = T_stratosphere_isa
        return T_isa

    # scalar
    return T_isa if h <= constants.h_tropopause else T_stratosphere_isa


def _low_altitude_m_to_pl(h: np.ndarray) -> np.ndarray:
    T_isa = _m_to_T_isa(h)
    power_term = -constants.g / (constants.T_lapse_rate * constants.R_d)
    return (constants.p_surface * (T_isa / constants.T_msl) ** power_term) / 100


def _high_altitude_m_to_pl(h: np.ndarray) -> np.ndarray:
    T_tropopause_isa = _m_to_T_isa(constants.h_tropopause)  # type: ignore[arg-type]
    power_term = -constants.g / (constants.T_lapse_rate * constants.R_d)
    p_tropopause_isa = constants.p_surface * (T_tropopause_isa / constants.T_msl) ** power_term
    inside_exp = (-constants.g / (constants.R_d * T_tropopause_isa)) * (h - constants.h_tropopause)
    return p_tropopause_isa * np.exp(inside_exp) / 100


def m_to_pl(h: np.ndarray) -> np.ndarray:
    r"""Convert from altitude (m) to pressure level (hPa).

    Parameters
    ----------
    h : ArrayLike
        altitude, [:math:`m`]

    Returns
    -------
    ArrayLike
        pressure level, [:math:`hPa`], [:math:`mbar`]

    Notes
    -----
    TODO: reference
    """
    # array
    if isinstance(h, (np.ndarray, xr.DataArray)):
        pl = _low_altitude_m_to_pl(h)
        pl[h > constants.h_tropopause] = _high_altitude_m_to_pl(h[h > constants.h_tropopause])
        return pl

    # scalar
    return _low_altitude_m_to_pl(h) if h <= constants.h_tropopause else _high_altitude_m_to_pl(h)


# ------
# SAC
# ------

jetA = JetA()


def T_critical_sac(
    air_temperature: ArrayLike,
    specific_humidity: ArrayLike,
    air_pressure: ArrayLike,
    engine_efficiency: float = 0.3,
    ei_h2o: float = jetA.ei_h2o,
    q_fuel: float = jetA.q_fuel,
    *,
    n_iter: int = 10,
    dx_threshold: float = 1e-3,
) -> ArrayLike:
    r"""Estimate actual Temperature threshold for persistent contrail formation iteratively.

    Applies Newton iteration and the Schmidt-Appleman criterion.

    This is required to estimate the activation rate of the aircraft emitted
    black carbon to contrail ice particles.

    Parameters
    ----------
    air_temperature : ArrayLike
        A sequence or array of temperature values, [:math:`K`]
    specific_humidity : ArrayLike
        A sequence or array of specific humidity values, [:math:`kg_{H_{2}O} \ kg_{air}`]
    air_pressure : ArrayLike
        A sequence or array of atmospheric pressure values, [:math:`Pa`].
    engine_efficiency: float, optional
        Engine efficiency, [:math:`0 - 1`]
    ei_h2o : float, optional
        Emission index of water vapor, [:math:`kg \ kg^{-1}`]
    q_fuel : float, optional
        Specific combustion heat of fuel combustion, [:math:`J \ kg^{-1} \ K^{-1}`]
    n_iter : int, optional
        Max number of iterations
    dx_threshold : float, optional
        Stop iteration once all elements of `dx` are smaller than `dx_threshold` in absolute value

    Returns
    -------
    ArrayLike
        Estimated Temperature threshold for contrail formation, [:math:`K`]

    Notes
    -----
    Source: Schumann, U., 1996. On conditions for contrail formation from aircraft exhausts.
            Meteorologische Zeitschrift, pp.4-23.
    """
    relative_humidity = thermo.rh(specific_humidity, air_temperature, air_pressure)
    G = sac.slope_mixing_line(specific_humidity, air_pressure, engine_efficiency, ei_h2o, q_fuel)
    T_sat_liquid_ = sac.T_sat_liquid(G)
    e_sat_T_sat_liquid = thermo.e_sat_liquid(T_sat_liquid_)

    # guess initial temperature
    relative_humidity_bounded = xr.where(relative_humidity < 1, relative_humidity, 0.9999)
    T0 = T_sat_liquid_ - (e_sat_T_sat_liquid / G)
    e_sat_T0 = thermo.e_sat_liquid(T0)
    c_1 = e_sat_T_sat_liquid * (T_sat_liquid_ - T0) / (2 * relative_humidity_bounded * e_sat_T0)
    c_2 = (
        1
        - (
            1
            + (4 * relative_humidity_bounded * e_sat_T0)
            / ((1 - relative_humidity_bounded) * e_sat_T_sat_liquid)
        )
        ** 0.5
    )
    T_crit_est = T_sat_liquid_ + (1 - relative_humidity) * c_1 * c_2

    # iterate up to n_iter until all elements of |dx| are smaller than dx_threshold
    for _ in np.arange(n_iter):
        e_sat_T_est = thermo.e_sat_liquid(T_crit_est)
        de_dt = (thermo.e_sat_liquid(T_crit_est + 1) - thermo.e_sat_liquid(T_crit_est - 1)) / 2
        dx = (
            e_sat_T_sat_liquid - G * (T_sat_liquid_ - T_crit_est) - relative_humidity * e_sat_T_est
        ) / (G - relative_humidity * de_dt)
        T_crit_est = T_crit_est - dx

        if (np.abs(dx) < dx_threshold).all():
            break

    # mask values where relative humidity is low
    is_dry = relative_humidity < 1e-6
    T_crit_est[is_dry] = T_sat_liquid_[is_dry] - (e_sat_T_sat_liquid[is_dry] / G[is_dry])

    # mask values where relative humidity is high
    is_liq_sat = relative_humidity > 0.9999
    T_crit_est[is_liq_sat] = T_sat_liquid_[is_liq_sat]

    return T_crit_est
